#
# CmbAccessibleEditor - Cambalache Accessible Editor
#
# Copyright (C) 2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .cmb_object import CmbObject
from .control import cmb_create_editor, CmbEnumComboBox
from .cmb_property_label import CmbPropertyLabel
from . import utils, _


class CmbAccessibleEditor(Gtk.Box):
    __gtype_name__ = "CmbAccessibleEditor"

    def __init__(self, **kwargs):
        self.__object = None
        self.__bindings = []
        self.__accessibility_metadata = None
        self.__role_filter_model = None

        super().__init__(**kwargs)

        self.props.orientation = Gtk.Orientation.VERTICAL

        self.__role_box = Gtk.Box(spacing=6)
        self.__role_box.append(Gtk.Label(label="accessible-role"))
        self.__role_combobox = CmbEnumComboBox()
        self.__role_box.append(self.__role_combobox)

        self.append(self.__role_box)

        self.__box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.append(self.__box)

    def bind_property(self, *args):
        binding = GObject.Object.bind_property(*args)
        self.__bindings.append(binding)
        return binding

    def __on_expander_expanded(self, expander, pspec, revealer):
        expanded = expander.props.expanded

        if expanded:
            revealer.props.transition_type = Gtk.RevealerTransitionType.SLIDE_DOWN
        else:
            revealer.props.transition_type = Gtk.RevealerTransitionType.SLIDE_UP

        revealer.props.reveal_child = expanded

    def __update_view(self):
        for child in utils.widget_get_children(self.__box):
            self.__box.remove(child)

        if self.__object is None or not self.__object.info.is_a("GtkWidget"):
            return

        obj = self.__object

        if obj.project.target_tk == "gtk-4.0":
            if not self.__object.info.is_a("GtkWidget"):
                self.__role_box.hide()
                return

            self.__role_box.show()

            prop = self.__object.properties_dict["accessible-role"]
            if prop.value in ["none", "presentation"]:
                # No need to set properties if role none or presentation
                return

            role_data = self.__object.project.db.accessibility_metadata.get(prop.value, None)
            if role_data:
                role_properties = role_data["properties"]
                role_states = role_data["states"]
            else:
                role_properties = None
                role_states = None

            a11y_data = [
                ("CmbAccessibleProperty", _("Properties"), len("cmb-a11y-properties"), role_properties),
                ("CmbAccessibleState", _("States"), len("cmb-a11y-states"), role_states),
                ("CmbAccessibleRelation", _("Relations"), None, None),
            ]
        else:
            type_actions = self.__object.project.db.accessibility_metadata.get(obj.type_id, [])

            a11y_data = [
                ("CmbAccessibleAction", _("Actions"), len("cmb-a11y-actions"), type_actions),
                ("CmbAccessibleProperty", _("Properties"), None, None),
                ("CmbAccessibleRelation", _("Relations"), None, None),
            ]
            self.__role_box.hide()

        properties = obj.properties_dict

        for owner_id, title, prefix_len, allowed_ids in a11y_data:
            info = obj.project.type_info.get(owner_id, None)

            if info is None:
                continue

            # Editor count
            i = 0

            # Grid for all editors
            grid = Gtk.Grid(hexpand=True, row_spacing=4, column_spacing=4)

            # Accessible iface properties
            for property_id in info.properties:
                # Ignore properties or status not for this role
                if prefix_len and allowed_ids is not None and property_id[prefix_len:] not in allowed_ids:
                    continue

                prop = properties.get(property_id, None)

                if prop is None or prop.info is None:
                    continue

                editor = cmb_create_editor(prop.project, prop.info.type_id, prop=prop)

                if editor is None:
                    return None, None

                self.bind_property(
                    prop,
                    "value",
                    editor,
                    "cmb-value",
                    GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
                )

                label = CmbPropertyLabel(prop=prop, bindable=False)
                grid.attach(label, 0, i, 1, 1)
                grid.attach(editor, 1, i, 1, 1)
                i += 1

            if i == 0:
                continue

            # Create expander/revealer to pack editor grid
            expander = Gtk.Expander(label=f"<b>{title}</b>", use_markup=True, expanded=True)
            revealer = Gtk.Revealer(reveal_child=True)
            expander.connect("notify::expanded", self.__on_expander_expanded, revealer)
            revealer.set_child(grid)
            self.__box.append(expander)
            self.__box.append(revealer)

        self.show()

    def __on_object_property_changed_notify(self, obj, prop, field):
        if prop.property_id == "accessible-role":
            self.__update_view()

    def __visible_func(self, model, iter, data):
        if self.__accessibility_metadata is None:
            return False

        name, nick, value = model[iter]

        role_data = self.__accessibility_metadata.get(nick, None)
        if role_data:
            # Ignore abstract roles
            if nick != "none" and role_data.get("is_abstract", False):
                return False

        return True

    @GObject.Property(type=CmbObject)
    def object(self):
        return self.__object

    @object.setter
    def _set_object(self, obj):
        if obj == self.__object:
            return

        if self.__object and self.__object.info.is_a("GtkWidget"):
            self.__object.disconnect_by_func(self.__on_object_property_changed_notify)
            self.__role_combobox.props.model = None
            self.__accessibility_metadata = None

        for binding in self.__bindings:
            binding.unbind()

        self.__bindings = []
        self.__object = obj

        if self.__object and self.__object.info.is_a("GtkWidget"):
            self.__object.connect("property-changed", self.__on_object_property_changed_notify)

            self.__accessibility_metadata = self.__object.project.db.accessibility_metadata
            a11y_info = self.__object.project.type_info.get("GtkAccessibleRole", None)

            if a11y_info:
                a11y_info = self.__object.project.type_info.get("GtkAccessibleRole", None)

                self.__role_filter_model = Gtk.TreeModelFilter(child_model=a11y_info.enum)
                self.__role_filter_model.set_visible_func(self.__visible_func)
                self.__role_combobox.info = a11y_info
                self.__role_combobox.props.model = self.__role_filter_model

                prop = self.__object.properties_dict.get("accessible-role", None)
                self.bind_property(
                        prop,
                        "value",
                        self.__role_combobox,
                        "cmb-value",
                        GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
                    )

        self.__update_view()


Gtk.WidgetClass.set_css_name(CmbAccessibleEditor, "CmbAccessibleEditor")



#
# CmbEnumComboBox
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk
from ..cmb_type_info import CmbTypeInfo


class CmbEnumComboBox(Gtk.ComboBox):
    __gtype_name__ = "CmbEnumComboBox"

    info = GObject.Property(type=CmbTypeInfo, flags=GObject.ParamFlags.READWRITE)
    text_column = GObject.Property(type=int, default=1, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.connect("changed", self.__on_changed)

        renderer_text = Gtk.CellRendererText()
        self.pack_start(renderer_text, True)
        self.add_attribute(renderer_text, "text", self.text_column)

        self.props.id_column = self.text_column

        if self.info:
            self.props.model = self.info.enum

    def __on_changed(self, obj):
        self.notify("cmb-value")

    @GObject.Property(type=str)
    def cmb_value(self):
        return self.props.active_id

    @cmb_value.setter
    def _set_cmb_value(self, value):
        if self.info is None:
            return

        self.props.active_id = None
        active_id = self.info.enum_get_value_as_string(value)

        if active_id == self.props.active_id:
            return

        self.props.active_id = active_id



#!/usr/bin/pytest

import json
import pytest
import threading
from unittest.mock import MagicMock
from uuid import uuid4

from cambalache.cmb_notification import notification_center, CmbNotification
from cambalache import utils
from . import utils as test_utils


DAY = 3600 * 24
now = utils.utcnow()
CMB_UUID = notification_center.uuid
POLL_UUID = str(uuid4())

POLL_NOTIFICATION_BASE = {
    "type": "poll",
    "start_date": now - DAY,
    "end_date":  now + DAY,
    "poll": {
        "start_date": now - DAY,
        "end_date": now + DAY,
        "allowed_votes": 1,
        "description": "A poll to test the frontend",
        "id": POLL_UUID,
        "options": [
            "First option",
            "Second option",
            "None of the above"
        ],
        "title": "Test Poll"
    }
}


def wait_for_all_threads():
    for t in threading.enumerate():
        if t == threading.current_thread():
            continue
        t.join()


def test_cmb_notification_disabled():
    assert notification_center.enabled is False
    assert len(notification_center.store) == 0
    notification_center.enabled = True


@pytest.mark.parametrize("response, n", [
    (
        {
            "notification": {
                "type": "version",
                "start_date": now - DAY,
                "end_date":  now + DAY,
                "version": "1.0.0",
                "release_notes": "Some Notes",
                "read_more_url": "http://localhost"
            }
        },
        1
    ),
    (
        {
            "notification": {
                "type": "message",
                "start_date": now - DAY,
                "end_date":  now + DAY,
                "title": "A simple message",
                "message": "This is a message notification"
            }
        },
        2
    ),
    (
        {
            "notification": POLL_NOTIFICATION_BASE
        },
        3
    ),
    (
        {
            "notification": {
                **POLL_NOTIFICATION_BASE,
                "results": {
                    "total": 146,
                    "votes": [12, 56, 78]
                }
            }
        },
        4
    )
])
def test_cmb_notification_get(mocker, response, n):
    wait_for_all_threads()

    mocker.patch(
        "http.client.HTTPSConnection.getresponse",
        return_value=MagicMock(
            status=200,
            read=MagicMock(return_value=json.dumps(response).encode())
        )
    )

    request_mock = mocker.patch("http.client.HTTPSConnection.request")

    notification_center.REQUEST_INTERVAL = 100
    notification_center.next_request = 0

    on_new_notification = MagicMock()
    notification_center.connect("new-notification", on_new_notification)

    notification_center._get_notification()

    test_utils.process_all_pending_gtk_events()

    request_mock.assert_called_with(
        "GET",
        "/notification",
        headers={
            'User-Agent': notification_center.user_agent,
            'x-cambalache-uuid': CMB_UUID
        }
    )

    assert notification_center.uuid == CMB_UUID
    on_new_notification.assert_called()

    assert len(notification_center.store) == n


def test_cmb_notification_get_interval(mocker):
    wait_for_all_threads()

    mocker.patch(
        "http.client.HTTPSConnection.getresponse",
        return_value=MagicMock(
            status=200,
            read=MagicMock(return_value=json.dumps({}).encode())
        )
    )

    request_mock = mocker.patch("http.client.HTTPSConnection.request")

    notification_center.REQUEST_INTERVAL = 1
    notification_center.next_request = 0

    notification_center._get_notification()

    # Test if more than one notification is called
    # First time will be called immediately and then every 1 second
    test_utils.process_all_pending_gtk_events(deciseconds=28)
    assert request_mock.call_count > 2
    assert notification_center.next_request > now

    notification_center.REQUEST_INTERVAL = 100
    notification_center.next_request = now + DAY


def test_cmb_notification_post_vote(mocker):
    wait_for_all_threads()

    poll = notification_center.store[0]
    assert isinstance(poll, CmbNotification)
    assert poll.poll.id == POLL_UUID
    assert poll.results.total == 146
    assert poll.results.votes == [12, 56, 78]

    mocker.patch(
        "http.client.HTTPSConnection.getresponse",
        return_value=MagicMock(
            status=200,
            read=MagicMock(return_value=json.dumps({
                "uuid": POLL_UUID,
                "results": {
                    "total": 1,
                    "votes": [0, 1, 0]
                }
            }).encode())
        )
    )

    request_mock = mocker.patch("http.client.HTTPSConnection.request")

    notification_center.poll_vote(poll, [1])

    test_utils.process_all_pending_gtk_events()

    request_mock.assert_called_with(
        "POST",
        f"/poll/{POLL_UUID}",
        json.dumps({"votes": [1]}),
        {
            "User-Agent": notification_center.user_agent,
            "x-cambalache-uuid": CMB_UUID,
            "Content-type": "application/json"
        }
    )

    assert poll.results.total == 1
    assert poll.results.votes == [0, 1, 0]


def test_cmb_notification_refresh_results(mocker):
    wait_for_all_threads()

    poll = notification_center.store[0]
    assert isinstance(poll, CmbNotification)
    assert poll.poll.id == POLL_UUID
    assert poll.results.total == 1
    assert poll.results.votes == [0, 1, 0]

    mocker.patch(
        "http.client.HTTPSConnection.getresponse",
        return_value=MagicMock(
            status=200,
            read=MagicMock(return_value=json.dumps({
                "uuid": POLL_UUID,
                "results": {
                    "total": 2,
                    "votes": [0, 1, 1]
                }
            }).encode())
        )
    )

    request_mock = mocker.patch("http.client.HTTPSConnection.request")

    notification_center.poll_refresh_results(poll)

    test_utils.process_all_pending_gtk_events()

    request_mock.assert_called_with(
        "GET",
        f"/poll/{POLL_UUID}",
        None,
        {
            "User-Agent": notification_center.user_agent,
            "x-cambalache-uuid": CMB_UUID,
            "Content-type": "application/json"
        }
    )

    assert poll.results.total == 2
    assert poll.results.votes == [0, 1, 1]


# This should cleanup all notifications
def test_cmb_notification_remove_all(mocker):
    wait_for_all_threads()

    assert notification_center.store.props.n_items > 0

    notification_center.remove_all()

    assert notification_center.store.props.n_items == 0



#
# Cambalache Object wrapper
#
# Copyright (C) 2021  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gio

from .cmb_list_error import CmbListError
from .cmb_base_objects import CmbBaseObject, CmbSignal
from .cmb_property import CmbProperty
from .cmb_layout_property import CmbLayoutProperty
from .cmb_object_data import CmbObjectData
from .cmb_type_info import CmbTypeInfo
from .cmb_ui import CmbUI
from .constants import GMENU_SECTION_TYPE,  GMENU_SUBMENU_TYPE, GMENU_ITEM_TYPE
from . import utils
from cambalache import getLogger, _

logger = getLogger(__name__)


class CmbObject(CmbBaseObject, Gio.ListModel):
    __gtype_name__ = "CmbObject"

    info = GObject.Property(type=CmbTypeInfo, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    __gsignals__ = {
        "property-changed": (GObject.SignalFlags.RUN_FIRST, None, (CmbProperty, str)),
        "layout-property-changed": (GObject.SignalFlags.RUN_FIRST, None, (GObject.GObject, CmbLayoutProperty)),
        "signal-added": (GObject.SignalFlags.RUN_FIRST, None, (CmbSignal,)),
        "signal-removed": (GObject.SignalFlags.RUN_FIRST, None, (CmbSignal,)),
        "signal-changed": (GObject.SignalFlags.RUN_FIRST, None, (CmbSignal,)),
        "data-added": (GObject.SignalFlags.RUN_FIRST, None, (CmbObjectData,)),
        "data-removed": (GObject.SignalFlags.RUN_FIRST, None, (CmbObjectData,)),
        "child-reordered": (GObject.SignalFlags.RUN_FIRST, None, (CmbBaseObject, int, int)),
    }

    def __init__(self, **kwargs):
        self.__properties = None
        self.__properties_dict = None
        self.__layout = None
        self.__layout_dict = None
        self.__signals = None
        self.__signals_dict = None
        self.__data = None
        self.__data_dict = None
        self.version_warning = None
        self.__is_template = False

        self._last_known = None

        super().__init__(**kwargs)

        self.connect("notify", self.__on_notify)

        if self.project is None:
            return

        self.__update_version_warning()

        self.connect("notify", self._on_notify)
        self.ui.connect("notify", self._on_ui_notify)
        self.ui.connect("library-changed", self._on_ui_library_changed)

    def __bool__(self):
        # Override Truth Value Testing to ensure that CmbObject objects evaluates to True even if it does not have children
        return True

    def __str__(self):
        return f"CmbObject<{self.display_name_type}> {self.ui_id}:{self.object_id}"

    @property
    def properties(self):
        self.__populate_properties()
        return self.__properties

    @property
    def properties_dict(self):
        self.__populate_properties()
        return self.__properties_dict

    @property
    def layout(self):
        self.__populate_layout()
        return self.__layout

    @property
    def layout_dict(self):
        self.__populate_layout()
        return self.__layout_dict

    @property
    def signals(self):
        self.__populate_signals()
        return self.__signals

    @property
    def signals_dict(self):
        self.__populate_signals()
        return self.__signals_dict

    @property
    def data(self):
        self.__populate_data()
        return self.__data

    @property
    def data_dict(self):
        self.__populate_data()
        return self.__data_dict

    @property
    def inline_property_id(self):
        ui_id = self.ui_id
        object_id = self.object_id
        parent_id = self.parent_id

        if parent_id:
            # Set which parent property makes a reference to this inline object
            row = self.project.db.execute(
                "SELECT property_id FROM object_property WHERE ui_id=? AND object_id=? AND inline_object_id=?;",
                (ui_id, parent_id, object_id)
            ).fetchone()
            return row[0] if row else None

        return None

    @property
    def binding_expression_property_id(self):
        ui_id = self.ui_id
        object_id = self.object_id
        parent_id = self.parent_id

        if parent_id:
            # Set which parent property makes a reference to this binding expression object
            row = self.project.db.execute(
                "SELECT property_id FROM object_property WHERE ui_id=? AND object_id=? AND binding_expression_id=?;",
                (ui_id, parent_id, object_id)
            ).fetchone()
            return row[0] if row else None

        return None

    def __populate_type_properties(self, name):
        property_info = self.project.get_type_properties(name)
        if property_info is None:
            return

        for property_name, info in property_info.items():
            # Check if this property was already installed by a derived class
            if property_name in self.__properties_dict:
                continue

            prop = CmbProperty(
                object=self,
                project=self.project,
                ui_id=self.ui_id,
                object_id=self.object_id,
                owner_id=name,
                property_id=info.property_id,
                info=info,
            )

            # List of property
            self.__properties.append(prop)

            # Dictionary of properties
            self.__properties_dict[property_name] = prop

    def __populate_properties(self):
        if self.__properties is not None:
            return
        self.__properties = []
        self.__properties_dict = {}

        self.__populate_type_properties(self.type_id)
        for parent_id in self.info.hierarchy:
            self.__populate_type_properties(parent_id)

            # Add accessible properties for GtkWidgets
            if parent_id == "GtkWidget":
                for accessible_id in [
                    "CmbAccessibleProperty",
                    "CmbAccessibleRelation",
                    "CmbAccessibleState",
                    "CmbAccessibleAction"
                ]:
                    self.__populate_type_properties(accessible_id)

    def __populate_layout_properties_from_type(self, name):
        property_info = self.project.get_type_properties(name)
        if property_info is None:
            return

        # parent_id is stored in the DB so its better to cache it
        parent_id = self.parent_id
        for property_name in property_info:
            info = property_info[property_name]

            prop = CmbLayoutProperty(
                object=self,
                project=self.project,
                ui_id=self.ui_id,
                object_id=parent_id,
                child_id=self.object_id,
                owner_id=name,
                property_id=info.property_id,
                info=info,
            )

            self.__layout.append(prop)

            # Dictionary of properties
            self.__layout_dict[property_name] = prop

    def _property_changed(self, prop, field):
        self.emit("property-changed", prop, field)
        self.project._object_property_changed(self, prop, field)

    def _layout_property_changed(self, prop):
        parent = self.project.get_object_by_id(self.ui_id, self.parent_id)
        self.emit("layout-property-changed", parent, prop)
        self.project._object_layout_property_changed(parent, self, prop)

    def __add_signal_object(self, signal):
        self.__populate_signals()
        self.__signals.append(signal)
        self.__signals_dict[signal.signal_pk] = signal
        self.emit("signal-added", signal)
        self.project._object_signal_added(self, signal)

        signal.connect("notify", self.__on_signal_notify)

    def __on_signal_notify(self, signal, pspec):
        self.emit("signal-changed", signal)
        self.project._object_signal_changed(self, signal)

    def __add_data_object(self, data):
        if data.get_id_string() in self.data_dict:
            return

        self.__data.append(data)
        self.__data_dict[data.get_id_string()] = data
        self.emit("data-added", data)
        self.project._object_data_added(self, data)

    def __on_notify(self, obj, pspec):
        if pspec.name == "parent-id":
            self.__populate_layout_properties()

        self.project._object_changed(self, pspec.name)

    def __populate_signals(self):
        if self.__signals is not None:
            return
        self.__signals = []
        self.__signals_dict = {}

        c = self.project.db.cursor()

        # Populate signals
        for row in c.execute("SELECT * FROM object_signal WHERE ui_id=? AND object_id=?;", (self.ui_id, self.object_id)):
            self.__add_signal_object(CmbSignal.from_row(self.project, *row))

    def __populate_data(self):
        if self.__data is not None:
            return
        self.__data = []
        self.__data_dict = {}

        c = self.project.db.cursor()

        # Populate data
        for row in c.execute(
            "SELECT * FROM object_data WHERE ui_id=? AND object_id=? AND parent_id IS NULL;",
            (self.ui_id, self.object_id),
        ):
            self.__add_data_object(CmbObjectData.from_row(self.project, *row))

    def __populate_layout_properties(self):
        parent_id = self.parent_id

        # FIXME: delete is anything is set?
        self.__layout = []
        self.__layout_dict = {}

        if parent_id > 0:
            parent = self.project.get_object_by_id(self.ui_id, parent_id)
            for owner_id in [parent.type_id] + parent.info.hierarchy:
                self.__populate_layout_properties_from_type(f"{owner_id}LayoutChild")

    def __populate_layout(self):
        if self.__layout is None:
            self.__populate_layout_properties()

    @GObject.Property(type=int)
    def parent_id(self):
        retval = self.db_get(
            "SELECT parent_id FROM object WHERE (ui_id, object_id) IS (?, ?);",
            (
                self.ui_id,
                self.object_id,
            ),
        )
        return retval if retval is not None else 0

    @parent_id.setter
    def _set_parent_id(self, value):
        new_parent_id = value if value != 0 else None
        old_parent_id = self.parent_id if self.parent_id != 0 else None

        if old_parent_id == new_parent_id:
            return

        # Save old parent and position
        self._save_last_known_parent_and_position()

        project = self.project
        ui_id = self.ui_id
        object_id = self.object_id

        if new_parent_id is None:
            new_position = self.db_get(
                "SELECT MAX(position)+1 FROM object WHERE ui_id=? AND parent_id IS NULL",
                (ui_id, )
            )
        else:
            new_position = self.db_get(
                "SELECT MAX(position)+1 FROM object WHERE ui_id=? AND parent_id=?",
                (ui_id, new_parent_id)
            )

        project.db.execute(
            "UPDATE object SET parent_id=?, position=? WHERE ui_id=? AND object_id=?;",
            (new_parent_id, new_position or 0, ui_id, object_id)
        )

        # Update children positions in old parent
        project.db.update_children_position(ui_id, old_parent_id)

        # Update GListModel
        self._remove_from_old_parent()
        self._update_new_parent()

        self.__populate_layout_properties()

    @GObject.Property(type=CmbUI)
    def ui(self):
        return self.project.get_object_by_id(self.ui_id)

    @GObject.Property(type=GObject.Object)
    def parent(self):
        return self.project.get_object_by_id(self.ui_id, self.parent_id)

    def _add_signal(self, signal_pk, owner_id, signal_id, handler, detail=None, user_data=0, swap=False, after=False):
        signal = CmbSignal(
            project=self.project,
            signal_pk=signal_pk,
            ui_id=self.ui_id,
            object_id=self.object_id,
            owner_id=owner_id,
            signal_id=signal_id,
            handler=handler,
            detail=detail,
            user_data=user_data if user_data is not None else 0,
            swap=swap,
            after=after,
        )

        self.__add_signal_object(signal)

        return signal

    def add_signal(self, owner_id, signal_id, handler, detail=None, user_data=0, swap=False, after=False):
        try:
            c = self.project.db.cursor()
            c.execute(
                """
                INSERT INTO object_signal (ui_id, object_id, owner_id, signal_id, handler, detail, user_data, swap, after)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?);
                """,
                (self.ui_id, self.object_id, owner_id, signal_id, handler, detail, user_data, swap, after),
            )
            signal_pk = c.lastrowid
            c.close()
            self.project.db.commit()
        except Exception as e:
            logger.warning(
                f"Error adding signal handler {owner_id}:{signal_id} {handler} to object {self.ui_id}.{{self.object_id}} {e}"
            )
            return None
        else:
            return self._add_signal(
                signal_pk,
                owner_id,
                signal_id,
                handler,
                detail=detail,
                user_data=user_data if user_data is not None else 0,
                swap=swap,
                after=after,
            )

    def _remove_signal(self, signal):
        self.__signals.remove(signal)
        del self.__signals_dict[signal.signal_pk]

        self.emit("signal-removed", signal)
        self.project._object_signal_removed(self, signal)

    def remove_signal(self, signal):
        try:
            self.project.db.execute("DELETE FROM object_signal WHERE signal_pk=?;", (signal.signal_pk,))
            self.project.db.commit()
        except Exception as e:
            handler = f"{signal.owner_id}:{signal.signal_id} {signal.handler}"
            logger.warning(f"Error removing signal handler {handler} from object {self.ui_id}.{{self.object_id}} {e}")
            return False
        else:
            self._remove_signal(signal)
            return True

    def _add_data(self, owner_id, data_id, id, info=None):
        data = CmbObjectData(
            project=self.project,
            object=self,
            info=info,
            ui_id=self.ui_id,
            object_id=self.object_id,
            owner_id=owner_id,
            data_id=data_id,
            id=id,
        )
        self.__add_data_object(data)
        return data

    def add_data(self, data_key, value=None, comment=None):
        try:
            value = str(value) if value is not None else None
            taginfo = self.info.get_data_info(data_key)
            owner_id = taginfo.owner_id
            data_id = taginfo.data_id
            id = self.project.db.object_add_data(self.ui_id, self.object_id, owner_id, data_id, value, None, comment)
        except Exception as e:
            logger.warning(f"Error adding data {data_key} {e}")
            return None
        else:
            return self._add_data(owner_id, data_id, id, info=taginfo)

    def _remove_data(self, data):
        if data.get_id_string() not in self.data_dict:
            return

        self.__data.remove(data)
        del self.__data_dict[data.get_id_string()]

        self.emit("data-removed", data)
        self.project._object_data_removed(self, data)

    def remove_data(self, data):
        try:
            assert data.get_id_string() in self.data_dict

            self.project.history_push(
                _("Remove {key} from {name}").format(key=data.info.key, name=self.display_name_type)
            )

            self.project.db.execute(
                "DELETE FROM object_data WHERE ui_id=? AND object_id=? AND owner_id=? AND data_id=? AND id=?;",
                (self.ui_id, self.object_id, data.owner_id, data.data_id, data.id),
            )
            self.project.db.commit()
            self.project.history_pop()
        except Exception as e:
            logger.warning(f"{self} Error removing data {data}: {e}")
            return False
        else:
            self._remove_data(data)
            return True

    def reorder_child(self, child, position):
        if child is None:
            logger.warning("child has to be a CmbObject")
            return

        if self.ui_id != child.ui_id or self.object_id != child.parent_id:
            logger.warning(f"{child} is not children of {self}")
            return

        old_position = child.position
        old_list_position = child.list_position
        if old_position == position:
            return

        name = child.name if child.name is not None else child.type_id
        self.project.history_push(
            _("Reorder object {name} from position {old} to {new}").format(name=name, old=old_position, new=position)
        )

        db = self.project.db

        # Consider this children
        #
        # label   0
        # button  1
        # entry   2
        # switch  3
        # toggle  4

        # Disable check so we can set position temporally to -1
        db.ignore_check_constraints = True
        db.execute("UPDATE object SET position=-1 WHERE ui_id=? AND object_id=?;", (self.ui_id, child.object_id))

        # Make room for new position
        for select_stmt, update_stmt in [
            (
                """
                SELECT ui_id, object_id
                FROM object
                WHERE ui_id=? AND parent_id=? AND position <= ? AND position > ?
                ORDER BY position ASC
                """,
                "UPDATE object SET position=position - 1 WHERE ui_id=? AND object_id=?;"
            ),
            (
                """
                SELECT ui_id, object_id
                FROM object
                WHERE ui_id=? AND parent_id=? AND position >= ? AND position < ?
                ORDER BY position DESC
                """,
                "UPDATE object SET position=position + 1 WHERE ui_id=? AND object_id=?;"
            ),
        ]:
            for row in db.execute(select_stmt, (self.ui_id, self.object_id, position, old_position)):
                db.execute(update_stmt, tuple(row))

        # Set new position
        db.execute("UPDATE object SET position=? WHERE ui_id=? AND object_id=?;", (position, self.ui_id, child.object_id))

        db.ignore_check_constraints = False

        list_position = child.list_position

        self.project._ignore_selection = True
        # Emit GListModel signals
        if position < old_position:
            self.items_changed(list_position, 0, 1)
            self.items_changed(old_list_position+1, 1, 0)
        else:
            if old_list_position != list_position:
                self.items_changed(old_list_position, 1, 0)
                self.items_changed(list_position, 0, 1)

        self.project._ignore_selection = False

        self.project.history_pop()
        self.emit("child-reordered", child, old_position, position)
        self.project._object_child_reordered(self, child, old_position, position)

    def clear_properties(self):
        c = self.project.db.cursor()

        name = self.name
        name = name if name is not None else self.type_id
        self.project.history_push(_("Clear object {name} properties").format(name=name))

        properties = []
        for row in c.execute(
            "SELECT property_id FROM object_property WHERE ui_id=? AND object_id=?;", (self.ui_id, self.object_id)
        ):
            properties.append(row[0])

        # Remove all properties from this object
        c.execute("DELETE FROM object_property WHERE ui_id=? AND object_id=?;", (self.ui_id, self.object_id))

        self.project.history_pop()
        c.close()

        for prop_id in properties:
            prop = self.__properties_dict[prop_id]
            prop.notify("value")
            self._property_changed(prop, prop_id)

    @GObject.Property(type=str)
    def display_name_type(self):
        return f"{self.type_id} {self.name}" if self.name else self.type_id

    @GObject.Property(type=str)
    def display_name(self):
        name = self.name or ""
        type_id = self.type_id
        parent_id = self.parent_id
        expression_property = self.binding_expression_property_id

        if type_id in [GMENU_SECTION_TYPE, GMENU_SUBMENU_TYPE, GMENU_ITEM_TYPE]:
            prop = self.properties_dict["label"]
            label = prop.value or ""
            display_name = f"{type_id} <i>{label}</i>"
        elif expression_property and type_id in ["GtkPropertyExpression", "GtkConstantExpression", "GtkClosureExpression"]:
            expression_type = {
                "GtkPropertyExpression": "lookup",
                "GtkConstantExpression": "constant",
                "GtkClosureExpression": "closure"
            }.get(type_id)
            display_name = f"<b>{expression_property}</b> ({expression_type} bind)"
        elif not parent_id and self.ui.template_id == self.object_id:
            # Translators: This is used for Template classes in the object tree
            display_name = _("{name} (template)").format(name=name)
        else:
            inline_prop = self.inline_property_id
            internal = self.internal
            if inline_prop:
                display_name = f"{type_id} <b>{inline_prop}</b> <i>{name}</i>"
            elif internal:
                display_name = f"{type_id} <b>{internal}</b> <i>{name}</i>"
            else:
                display_name = f"{type_id} <i>{name}</i>"

        if self.version_warning:
            return f'<span underline="error">{display_name}</span>'
        else:
            return display_name

    def __update_version_warning(self):
        target = self.ui.get_target(self.info.library_id)
        self.version_warning = utils.get_version_warning(target, self.info.version, self.info.deprecated_version, self.type_id)

    def _on_notify(self, obj, pspec):
        property_id = pspec.name

        if property_id in ["name", "type-id"]:
            self.notify("display-name")
            self.notify("display-name-type")

    def _on_ui_notify(self, obj, pspec):
        property_id = pspec.name

        if property_id == "template-id":
            was_template = self.__is_template
            self.__is_template = obj.template_id == self.object_id

            if was_template or self.__is_template:
                self.notify("display-name")
                self.notify("display-name-type")

    def _on_ui_library_changed(self, ui, library_id):
        self.__update_version_warning()

        self.__populate_properties()
        self.__populate_layout()

        # Update properties directly, to avoid having to connect too many times to this signal
        for props in [self.__properties, self.__layout]:
            for prop in props:
                if prop.library_id == library_id:
                    prop._update_version_warning()

    # GListModel helpers
    def _save_last_known_parent_and_position(self):
        self._last_known = (self.parent, self.list_position)

    def _update_new_parent(self):
        parent = self.parent
        position = self.list_position

        # Emit GListModel signal to update model
        if parent:
            parent.items_changed(position, 0, 1)
            parent.notify("n-items")
        else:
            ui = self.ui
            ui.items_changed(position, 0, 1)
            ui.notify("n-items")

        self._last_known = None

    def _remove_from_old_parent(self):
        if self._last_known is None:
            return

        parent, position = self._last_known

        # Emit GListModel signal to update model
        if parent:
            parent.items_changed(position, 1, 0)
            parent.notify("n-items")
        else:
            ui = self.ui
            ui.items_changed(position, 1, 0)
            ui.notify("n-items")

        self._last_known = None

    @GObject.Property(type=int)
    def list_position(self):
        ui_id = self.ui_id

        if self.parent_id:
            retval = self.db_get(
                """
                SELECT rownum-1
                FROM (
                    SELECT ROW_NUMBER() OVER (ORDER BY position ASC) rownum, object_id
                    FROM object
                    WHERE ui_id=? AND parent_id=?
                )
                WHERE object_id=?;
                """,
                (ui_id, self.parent_id, self.object_id)
            )
        else:
            retval = self.db_get(
                """
                SELECT rownum-1
                FROM (
                    SELECT ROW_NUMBER() OVER (ORDER BY position ASC) rownum, object_id
                    FROM object
                    WHERE ui_id=? AND parent_id IS NULL
                )
                WHERE object_id=?;
                """,
                (ui_id, self.object_id)
            )

        return retval

    # GListModel iface
    def do_get_item(self, position):
        ui_id = self.ui_id

        # This query should use index object_ui_id_parent_id_position_idx
        retval = self.db_get(
            """
            SELECT object_id
            FROM (
                SELECT ROW_NUMBER() OVER (ORDER BY position ASC) rownum, object_id
                FROM object
                WHERE ui_id=? AND parent_id=?
            )
            WHERE rownum=?;
            """,
            (ui_id, self.object_id, position+1)
        )
        if retval is not None:
            return self.project.get_object_by_id(ui_id, retval)

        # This should not happen
        return CmbListError()

    def do_get_item_type(self):
        return CmbBaseObject

    @GObject.Property(type=int)
    def n_items(self):
        if self.project is None:
            return 0

        retval = self.db_get("SELECT COUNT(object_id) FROM object WHERE ui_id=? AND parent_id=?;", (self.ui_id, self.object_id))
        return retval if retval is not None else 0

    def do_get_n_items(self):
        return self.n_items



#
# CmbMessageNotificationView
#
# Copyright (C) 2025  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from cambalache import getLogger
from gi.repository import GObject, Gtk
from .cmb_notification import CmbMessageNotification

logger = getLogger(__name__)


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_message_notification_view.ui")
class CmbMessageNotificationView(Gtk.Box):
    __gtype_name__ = "CmbMessageNotificationView"

    notification = GObject.Property(
        type=CmbMessageNotification, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY
    )

    # Message
    title_label = Gtk.Template.Child()
    message_label = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        notification = self.notification
        self.title_label.props.label = f"<b>{notification.title}</b>"
        self.message_label.props.label = notification.message



#
# CmbDBInspector
#
# Copyright (C) 2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GLib, GObject, Gio, Gtk
from cambalache import CmbProject


class CmbDBTable(GObject.Object):
    def __init__(self, **kwargs):
        self.__properties = {}
        super().__init__(**kwargs)

    def do_get_property(self, prop):
        # TODO: read from DB directly
        if not prop.name.startswith("cmb-int-") and prop.name not in self.__properties_set__:
            raise AttributeError('unknown property %s' % prop.name)
        return self.__properties[prop.name]

    def do_set_property(self, prop, value):
        # TODO: only store PK values when using DB
        if not prop.name.startswith("cmb-int-") and prop.name not in self.__properties_set__:
            raise AttributeError('unknown property %s' % prop.name)
        self.__properties[prop.name] = value
        self.notify(prop.name)


class CmbDBStore(GObject.GObject, Gio.ListModel):
    __gtype_name__ = 'CmbDBStore'

    project = GObject.Property(type=CmbProject, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    def __init__(self, ItemClass, **kwargs):
        super().__init__(**kwargs)

        self.__item_class = ItemClass
        self.__history_index = None
        self._objects = []

    def do_get_item(self, position):
        self.__check_refresh()
        return self._objects[position] if position < len(self._objects) else None

    def do_get_item_type(self):
        return self.__item_class

    def do_get_n_items(self):
        self.__check_refresh()
        return len(self._objects)

    def __check_refresh(self):
        history_index = self.project.history_index

        # Nothing to update if history did not changed
        if history_index == self.__history_index:
            return

        ItemClass = self.__item_class
        properties = ItemClass.__properties__
        int_properties = ItemClass.__int_properties__
        table = ItemClass.__table__
        needs_update = False

        # Basic optimization, only update if something changed in this table
        # TODO: this could be optimized more by check command to know exactly which row changed
        if self.__history_index is None or table.startswith("history") or table in ["global", "__profile__"]:
            needs_update = True
        else:
            change_table = table[7:] if table.startswith("history_") else table

            # TODO: detect command compression
            for row in self.project.db.execute(
                "SELECT table_name FROM history WHERE history_id >= ? ORDER BY history_id;", (self.__history_index, )
            ):
                table_name, = row
                if table_name == change_table:
                    needs_update = True
                    break

        self.__history_index = history_index

        if not needs_update:
            return

        # Emit signal to clear model
        n_items = len(self._objects)
        if n_items:
            self._objects = []
            self.items_changed(0, n_items, 0)

        if len(ItemClass.__pk__):
            pk_columns = ",".join(ItemClass.__pk__)
        else:
            pk_columns = "rowid"

        for row in self.project.db.execute(f"SELECT * FROM {table} ORDER BY {pk_columns};"):
            item = ItemClass()
            for i, val in enumerate(row):
                property_id = properties[i]
                if property_id in int_properties:
                    item.set_property(f"cmb-int-{property_id}", val if val is not None else 0)
                item.set_property(property_id, val)

            self._objects.append(item)

        # Emit signal to populate model
        self.items_changed(0, 0, len(self._objects))


class TableView(Gtk.ColumnView):
    project = GObject.Property(type=CmbProject, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    def __init__(self, ItemClass, **kwargs):
        super().__init__(**kwargs)

        self.props.show_row_separators = True
        self.props.show_column_separators = True
        self.props.reorderable = False
        self.__model = None
        self.__filter_model = None
        self.__item_class = ItemClass

        for property_id in ItemClass.__properties__:
            factory = Gtk.SignalListItemFactory()
            factory.connect("setup", self._on_factory_setup)
            factory.connect("bind", self._on_factory_bind, property_id)
            factory.connect("unbind", self._on_factory_unbind)

            col = Gtk.ColumnViewColumn(title=property_id, factory=factory)

            if property_id in ItemClass.__int_properties__:
                property_expression = Gtk.PropertyExpression.new(ItemClass, None, f"cmb-int-{property_id}")
                sorter = Gtk.NumericSorter()
            else:
                property_expression = Gtk.PropertyExpression.new(ItemClass, None, property_id)
                sorter = Gtk.StringSorter()
                col.props.resizable = True
                col.props.expand = True

            sorter.set_expression(property_expression)
            col.set_sorter(sorter)
            self.append_column(col)

        # TODO: keep track of project changes only while we are showing this model
        self.connect("map", self.__on_map)
        self.project.connect("changed", self.__on_project_changed)

    def __update_label(self, item, label, property_id):
        val = str(item.get_property(property_id))
        label.set_text(val if val else "")

    def __on_item_notify(self, item, pspec, label):
        self.__update_label(item, label, pspec.name)

    def _on_factory_setup(self, factory, list_item):
        label = Gtk.Inscription()
        list_item.set_child(label)

    def _on_factory_bind(self, factory, list_item, property_id):
        label = list_item.get_child()
        item = list_item.get_item()

        self.__update_label(item, label, property_id)
        item.connect(f"notify::{property_id}", self.__on_item_notify, label)

    def _on_factory_unbind(self, factory, list_item):
        item = list_item.get_item()
        item.disconnect_by_func(self.__on_item_notify)

    def __on_map(self, w):
        # Trigger check refresh
        if self.__model is not None:
            self.__model.get_n_items()
            return

        # Load model when widget is shown
        self.__model = CmbDBStore(self.__item_class, project=self.project)
        self.__filter_model = Gtk.SortListModel(model=self.__model, sorter=self.get_sorter())
        self.set_model(Gtk.NoSelection(model=self.__filter_model))

    def __on_project_changed(self, project):
        # Trigger check refresh
        if self.__model is not None and self.is_visible():
            self.__model.get_n_items()


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_db_inspector.ui")
class CmbDBInspector(Gtk.Box):
    __gtype_name__ = "CmbDBInspector"

    stack = Gtk.Template.Child()

    def __init__(self, **kwargs):
        self.__project = None
        self.__table_classes = None
        super().__init__(**kwargs)

        self.connect("map", self.__on_map)

    @GObject.Property(type=CmbProject)
    def project(self):
        return self.__project

    @project.setter
    def _set_project(self, project):
        self.__project = project

    def __init_tables(self):
        db = self.project.db
        self.__table_classes = {}

        for row in db.execute("SELECT name FROM sqlite_master WHERE type='table' ORDER BY name;"):
            table, = row

            if table.startswith("sqlite_"):
                continue

            klass = self.__class_from_table(table)
            self.__table_classes[table] = klass

    def _metadata_from_table(self, table):
        db = self.project.db
        properties = []
        gproperties = {}
        int_properties = set()
        pk_list = []

        for row in db.execute(f"PRAGMA table_info({table});"):
            col = row[1]
            col_type = row[2]
            pk = row[5]

            name = col.replace("_", "-")
            properties.append(name)
            if col_type == "INTEGER":
                int_properties.add(name)
                gproperties[f"cmb-int-{name}"] = (int, "", "", GLib.MININT, GLib.MAXINT, 0, GObject.ParamFlags.READWRITE)

            gproperties[name] = (str, "", "", None, GObject.ParamFlags.READWRITE)

            if pk:
                pk_list.append(col)

        return properties, gproperties, int_properties, pk_list

    def __class_from_table(self, table):
        class_name = f"CmbDBTable_{table}"
        properties, gproperties, int_properties, pk = self._metadata_from_table(table)
        klass = type(class_name, (CmbDBTable,), dict(
            __table__=table,
            __gproperties__=gproperties,
            __properties__=properties,
            __properties_set__=set(properties),
            __int_properties__=int_properties,
            __pk__=pk)
        )
        return klass

    def __populate_stack(self):
        for table, klass in self.__table_classes.items():
            sw = Gtk.ScrolledWindow(
                hexpand=True,
                vexpand=True,
                propagate_natural_width=True,
                propagate_natural_height=True)
            view = TableView(klass, project=self.__project)
            sw.set_child(view)
            self.stack.add_titled(sw, table, table)

    def __on_map(self, w):
        if self.__table_classes is None and self.__project is not None:
            self.__init_tables()
            self.__populate_stack()



#
# CmbView - Cambalache View
#
# Copyright (C) 2021-2026  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os
import json
import time
import atexit
import socket

from . import config
from .cmb_ui import CmbUI
from .cmb_object import CmbObject
from .cmb_context_menu import CmbContextMenu

from gi.repository import GObject, GLib, Gtk, Casilda

from cambalache import getLogger, _, ngettext
from cambalache.cmb_blueprint import cmb_blueprint_decompile
from mrg_command import MrgCommand


logger = getLogger(__name__)

basedir = os.path.dirname(__file__) or "."

GObject.type_ensure(Casilda.Compositor.__gtype__)


class CmbMerengueProcess(GObject.Object, MrgCommand):
    __gsignals__ = {
        "handle-command": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "exit": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    gtk_version = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    compositor = GObject.Property(type=GObject.Object, flags=GObject.ParamFlags.READWRITE)
    gresource_overlays = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        self.__file = os.path.join(config.merenguedir, "merengue", "merengue")
        self.__server_socket = None
        self.__pid = 0

        self.init_command()

        super().__init__(**kwargs)

    @GObject.Property(type=int)
    def pid(self):
        return self.__pid

    def cleanup(self):
        self.stop()

    def handle_command(self, line):
        self.emit("handle-command", line)

    def start(self):
        if self.__file is None or self.__pid > 0:
            return

        # Create socketpair for commands comunication
        client, server = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM)

        # Keep a reference in python to avoid socket close
        self.__server_socket = server

        # Init connection
        self.init_connection(server.fileno())

        # Get a socket already connected to the compositor
        wayland_socket = self.compositor.get_client_socket_fd()

        # Create Merengue environment, dont include variables that interfere with backend choice
        ignore_vars = {"GDK_BACKEND", "WAYLAND_SOCKET", "WAYLAND_DISPLAY"}

        envp = [f"{var}={val}" for var, val in os.environ.items() if var not in ignore_vars]

        # Force Gdk backend to wayland
        envp.append("GDK_BACKEND=wayland")

        # This is used for merengue to map resource files without compiling and loading a gresource bundle
        if self.gresource_overlays:
            envp.append(f"G_RESOURCE_OVERLAYS={self.gresource_overlays}")

        # Use WAYLAND_SOCKET instead of WAYLAND_DISPLAY
        envp.append(f"WAYLAND_SOCKET={wayland_socket}")

        if "MERENGUE_FORCE_SOFTWARE" in os.environ:
            envp.append("CASILDA_FORCE_SOFTWARE=1")
            envp.append("GSK_RENDERER=cairo")

        # Spawn merengue with wayland_socket and client file descriptors
        # TODO: send stdout and error to some place useful for the user
        valid, pid, stdin, stdout, stderr = GLib.spawn_async_with_pipes_and_fds(
            None,
            [self.__file, self.gtk_version, str(client.fileno())],
            envp,
            GLib.SpawnFlags.DO_NOT_REAP_CHILD |
            GLib.SpawnFlags.CHILD_INHERITS_STDIN |
            GLib.SpawnFlags.CHILD_INHERITS_STDERR |
            GLib.SpawnFlags.CHILD_INHERITS_STDOUT,
            None, None,
            -1, -1, -1,
            [wayland_socket, client.fileno()],
            [wayland_socket, client.fileno()]
        )

        # Close file descriptors passed to the client.
        # If we do not close wayland_socket then the client wont be destroyed if its process is killed.
        os.close(wayland_socket)
        client.close()

        if valid:
            self.__pid = pid
            GLib.child_watch_add(GLib.PRIORITY_DEFAULT_IDLE, pid, self.__on_exit, None)

    def close_connection(self):
        super().close_connection()
        self.__server_socket = None

    def stop(self):
        self.close_connection()

        if self.__pid:
            try:
                GLib.spawn_close_pid(self.__pid)
                os.kill(self.__pid, 9)
            except Exception as e:
                logger.warning(f"Error stopping {self.__file} {e}")
            finally:
                self.__pid = 0

    def __on_exit(self, pid, status, data):
        logger.warning("Merengue process exited")
        self.close_connection()
        self.__pid = 0
        self.emit("exit")


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_view.ui")
class CmbView(Gtk.Box):
    __gtype_name__ = "CmbView"

    __gsignals__ = {
        "placeholder-selected": (GObject.SignalFlags.RUN_LAST, None, (int, int, object, int, str)),
        "placeholder-activated": (GObject.SignalFlags.RUN_LAST, None, (int, int, object, int, str)),
    }

    show_merengue = GObject.Property(type=bool, default=False, flags=GObject.ParamFlags.READWRITE)
    preview = GObject.Property(type=bool, default=False, flags=GObject.ParamFlags.READWRITE)

    stack = Gtk.Template.Child()
    compositor = Gtk.Template.Child()
    compositor_offload = Gtk.Template.Child()
    compositor_box = Gtk.Template.Child()
    error_box = Gtk.Template.Child()
    error_message = Gtk.Template.Child()
    text_view = Gtk.Template.Child()
    db_inspector = Gtk.Template.Child()

    def __init__(self, **kwargs):
        self.__project = None
        self.__ui = None
        self.__theme = None
        self.__css_update_timeout = {}

        self.menu = self.__create_context_menu()

        super().__init__(**kwargs)

        self.__click_gesture = Gtk.GestureClick(propagation_phase=Gtk.PropagationPhase.CAPTURE, button=3)
        self.__click_gesture.connect("released", self.__on_click_gesture_released)
        self.compositor_box.add_controller(self.__click_gesture)

        self.__merengue = CmbMerengueProcess(compositor=self.compositor)
        self.__merengue.connect("exit", self.__on_process_exit)
        self.__merengue_last_exit = None
        self.__merengue_started = None
        self.__restart_merenge_timeout_source = None

        self.connect("notify::preview", self.__on_preview_notify)

        # Ensure we delete all socket files when exiting
        atexit.register(self.__atexit)

    @Gtk.Template.Callback("on_restart_button_clicked")
    def __on_restart_button_clicked(self, button):
        self.restart_workspace()

    def __atexit(self):
        self.__merengue_command("quit")
        self.__merengue.cleanup()

    def __merengue_command(self, command, args=None):
        if not self.__merengue_started:
            return

        self.__merengue.write_command(command, args)

    def __get_ui_xml(self, ui_id, merengue=False):
        if self.show_merengue:
            merengue = True

        return self.__project.db.tostring(ui_id, merengue=merengue)

    def __update_view(self):
        if self.__project and self.__ui:
            if self.stack.props.visible_child_name == "ui_xml":
                ui_source = self.__get_ui_xml(self.__ui.ui_id)

                if self.__ui.filename and self.__ui.filename.endswith(".blp"):
                    try:
                        ui_source = cmb_blueprint_decompile(ui_source)
                        self.text_view.lang = "blueprint"
                    except Exception as e:
                        ui_source = _("Error exporting project")
                        ui_source += "\n"
                        ui_source += ngettext(
                            "blueprintcompiler encounter the following error:",
                            "blueprintcompiler encounter the following errors:",
                            len(e.errors)
                        )
                        ui_source += "\n"
                        ui_source += str(e)
                        self.text_view.lang = ""
                        # TODO: forward error to parent to show to user
                else:
                    self.text_view.lang = "xml"

                self.text_view.buffer.set_text(ui_source)
            return

        self.text_view.buffer.set_text("")
        self.__ui = None

    def __get_ui_dirname(self, ui_id):
        dirname = GLib.get_home_dir()

        # Use project dir as default base directory
        if self.__project.filename:
            dirname = os.path.dirname(self.__project.filename)
        else:
            dirname = os.getcwd()

        # Use UI directory
        ui = self.__project.get_object_by_id(ui_id)
        if ui and ui.filename:
            dirname = os.path.join(dirname, os.path.dirname(ui.filename))

        return dirname

    def __merengue_update_ui(self, ui_id):
        ui = self.__get_ui_xml(ui_id, merengue=True) if ui_id else None
        toplevels = self.__project.db.get_toplevels(ui_id)
        selection = self.__project.get_selection()
        objects = self.__get_selection_objects(selection, ui_id)

        self.__merengue_command(
            "update_ui",
            args={
                "ui_id": ui_id,
                "dirname": self.__get_ui_dirname(ui_id),
                "toplevels": toplevels,
                "selection": objects,
                "xml": ui,
            },
        )

    def __restart_merenge_timeout(self, data):
        self.restart_workspace()
        self.__restart_merenge_timeout_source = 0
        return GLib.SOURCE_REMOVE

    def __on_notify(self, project, pspec):
        if pspec.name == "gresource-overlays":
            self.__merengue.gresource_overlays = project.gresource_overlays
            # Limit restart to 4 seconds
            if self.__restart_merenge_timeout_source:
                GLib.source_remove(self.__restart_merenge_timeout_source)

            self.__restart_merenge_timeout_source = GLib.timeout_add_seconds(4, self.__restart_merenge_timeout, None)

    def __on_changed(self, project):
        self.__update_view()

    def __on_ui_changed(self, project, ui, field):
        if field in ["custom-fragment", "filename"]:
            self.__merengue_update_ui(ui.ui_id)

    def __on_object_added(self, project, obj):
        self.__merengue_update_ui(obj.ui_id)

    def __on_object_removed(self, project, obj):
        self.__merengue_update_ui(obj.ui_id)

    def __on_object_changed(self, project, obj, field):
        if field in ["type", "position", "custom-fragment", "parent-id"]:
            self.__merengue_update_ui(obj.ui_id)

    def __on_object_property_changed(self, project, obj, prop, field):
        if field != "value":
            return

        info = prop.info

        # FIXME: implement new merengue command for updating a11y props
        if info.is_a11y:
            return

        if obj.info.workspace_type is None and info.construct_only:
            self.__merengue_update_ui(obj.ui_id)
            return

        self.__merengue_command(
            "object_property_changed",
            args={
                "ui_id": obj.ui_id,
                "object_id": obj.object_id,
                "property_id": prop.property_id,
                "is_object": prop.info.is_object,
                "value": prop.value,
            },
        )

    def __on_object_layout_property_changed(self, project, obj, child, prop):
        self.__merengue_command(
            "object_layout_property_changed",
            args={
                "ui_id": obj.ui_id,
                "object_id": obj.object_id,
                "child_id": child.object_id,
                "property_id": prop.property_id,
                "value": prop.value,
            },
        )

    def __on_object_property_binding_changed(self, project, obj, prop):
        self.__merengue_update_ui(obj.ui_id)

    def __get_selection_objects(self, selection, ui_id):
        objects = []

        for obj in selection:
            if type(obj) is CmbObject and obj.ui_id == ui_id:
                objects.append(obj.object_id)

        return objects

    def __on_project_selection_changed(self, project):
        selection = project.get_selection()

        if len(selection) > 0:
            obj = selection[0]

            if isinstance(obj, CmbUI):
                ui = obj
            elif isinstance(obj, CmbObject):
                ui = obj.ui
            else:
                return

            ui_id = obj.ui_id

            if self.__ui != ui:
                self.__ui = ui
                self.__merengue_update_ui(ui.ui_id)

            objects = self.__get_selection_objects(selection, ui.ui_id)
            self.__merengue_command("selection_changed", args={"ui_id": ui_id, "selection": objects})
        else:
            self.__ui = None
            self.__merengue_update_ui(0)

        self.__update_view()

    def __on_css_added(self, project, obj):
        self.__merengue_command(
            "add_css_provider",
            args={
                "css_id": obj.css_id,
                "css": obj.css,
                "priority": obj.priority,
                "is_global": obj.is_global,
                "provider_for": obj.provider_for,
            },
        )

    def __on_css_removed(self, project, obj):
        self.__merengue_command("remove_css_provider", args={"css_id": obj.css_id})

    def __update_css_timeout(self, data):
        self.__merengue_command("update_css_provider", args=data)
        self.__css_update_timeout.pop(data["css_id"])
        return GLib.SOURCE_REMOVE

    def __on_css_changed(self, project, obj, field):
        data = {
            "css_id": obj.css_id,
            "field": field,
            "value": obj.get_property(field)
        }

        # Limit rate to 1 seconds
        if field == "css":
            last_timeout = self.__css_update_timeout.get(obj.css_id, None)

            # Remove old timeout for this CSS
            if last_timeout is not None:
                GLib.source_remove(last_timeout)

            # Queue CSS update
            self.__css_update_timeout[obj.css_id] = GLib.timeout_add_seconds(1, self.__update_css_timeout, data)
        else:
            self.__merengue_command("update_css_provider", args=data)

    def __on_library_info_changed(self, project, info, field):
        if field != "enabled":
            return

        if info.enabled:
            self.__merengue_command(
                "load_namespace",
                args={
                    "namespace": info.namespace,
                    "version": info.version,
                    "object_types": info.object_types,
                },
            )
        else:
            # Restart workspace so unused namespaces are unloaded
            self.restart_workspace()

    def __on_object_data_changed(self, project, data):
        self.__merengue_update_ui(data.ui_id)

    def __on_object_data_removed(self, project, obj, data):
        self.__merengue_update_ui(data.ui_id)

    def __on_object_data_data_removed(self, project, parent, data):
        self.__merengue_update_ui(data.ui_id)

    def __on_object_data_arg_changed(self, project, data, value):
        self.__merengue_update_ui(data.ui_id)

    def __on_object_child_reordered(self, project, obj, child, old_position, new_position):
        self.__merengue_update_ui(obj.ui_id)

    def __set_error_message(self, message):
        if message:
            self.error_message.props.label = message
            self.compositor_offload.set_visible(False)
            self.error_box.set_visible(True)
        else:
            self.error_message.props.label = ""
            self.compositor_offload.set_visible(True)
            self.error_box.set_visible(False)

    @GObject.Property(type=GObject.GObject)
    def project(self):
        return self.__project

    @project.setter
    def _set_project(self, project):
        if self.__project:
            self.__project.disconnect_by_func(self.__on_notify)
            self.__project.disconnect_by_func(self.__on_changed)
            self.__project.disconnect_by_func(self.__on_ui_changed)
            self.__project.disconnect_by_func(self.__on_object_added)
            self.__project.disconnect_by_func(self.__on_object_removed)
            self.__project.disconnect_by_func(self.__on_object_changed)
            self.__project.disconnect_by_func(self.__on_object_property_changed)
            self.__project.disconnect_by_func(self.__on_object_layout_property_changed)
            self.__project.disconnect_by_func(self.__on_object_property_binding_changed)
            self.__project.disconnect_by_func(self.__on_object_data_changed)
            self.__project.disconnect_by_func(self.__on_object_data_removed)
            self.__project.disconnect_by_func(self.__on_object_data_data_removed)
            self.__project.disconnect_by_func(self.__on_object_data_arg_changed)
            self.__project.disconnect_by_func(self.__on_object_child_reordered)
            self.__project.disconnect_by_func(self.__on_project_selection_changed)
            self.__project.disconnect_by_func(self.__on_css_added)
            self.__project.disconnect_by_func(self.__on_css_removed)
            self.__project.disconnect_by_func(self.__on_css_changed)
            self.__project.disconnect_by_func(self.__on_library_info_changed)
            self.__merengue.disconnect_by_func(self.__on_merengue_handle_command)
            self.__merengue.stop()

        self.__project = project
        self.db_inspector.project = project

        if project:
            project.connect("notify", self.__on_notify)
            project.connect("changed", self.__on_changed)
            project.connect("ui-changed", self.__on_ui_changed)
            project.connect("object-added", self.__on_object_added)
            project.connect("object-removed", self.__on_object_removed)
            project.connect("object-changed", self.__on_object_changed)
            project.connect("object-property-changed", self.__on_object_property_changed)
            project.connect("object-layout-property-changed", self.__on_object_layout_property_changed)
            project.connect("object-property-binding-changed", self.__on_object_property_binding_changed)
            project.connect("object-data-changed", self.__on_object_data_changed)
            project.connect("object-data-removed", self.__on_object_data_removed)
            project.connect("object-data-data-removed", self.__on_object_data_data_removed)
            project.connect("object-data-arg-changed", self.__on_object_data_arg_changed)
            project.connect("object-child-reordered", self.__on_object_child_reordered)
            project.connect("selection-changed", self.__on_project_selection_changed)
            project.connect("css-added", self.__on_css_added)
            project.connect("css-removed", self.__on_css_removed)
            project.connect("css-changed", self.__on_css_changed)
            project.connect("library-info-changed", self.__on_library_info_changed)
            self.__merengue.connect("handle-command", self.__on_merengue_handle_command)

            # Run view process
            if project.target_tk == "gtk+-3.0":
                self.__merengue.gtk_version = "3.0"
            elif project.target_tk == "gtk-4.0":
                self.__merengue.gtk_version = "4.0"

            # Sync GResource overlays
            self.__merengue.gresource_overlays = project.gresource_overlays

            # Clear any error
            self.__set_error_message(None)
            self.__merengue.start()

            # Update css themes
            self.menu.target_tk = project.target_tk

        self.__update_view()

    @GObject.Property(type=str)
    def gtk_theme(self):
        return self.__theme

    @gtk_theme.setter
    def _set_theme(self, theme):
        self.__theme = theme
        self.__merengue_command("gtk_settings_set", args={"property": "gtk-theme-name", "value": theme})

    def __on_click_gesture_released(self, gesture, n_press, x, y):
        if n_press == 1 and gesture.get_current_button() == 3 and \
         x >= 0 and x <= self.compositor_box.get_width() and y >= 0 and y <= self.compositor_box.get_height():
            self.menu.popup_at(x, y)

    def inspect(self):
        self.stack.props.visible_child_name = "ui_xml"
        self.__update_view()

    def set_interactive_debugging(self, enable):
        self.__merengue_command("set_interactive_debugging", args={"enable": enable})

    def restart_workspace(self):
        # Clear last exit timestamp
        self.__merengue_last_exit = None
        self.__merengue_started = None

        if self.__merengue.pid:
            # Let __on_process_exit() restart Merengue
            self.__merengue.stop()
        else:
            self.__set_error_message(None)
            self.__merengue.start()

    def __create_context_menu(self):
        retval = CmbContextMenu(enable_theme=True)
        retval.set_parent(self)

        retval.main_section.append(_("Restart workspace"), "win.workspace_restart")
        retval.main_section.append(_("Inspect UI definition"), "win.inspect")
        retval.main_section.append(_("Gtk Inspector"), "win.open_inspector")

        return retval

    def __on_process_exit(self, process):
        self.__merengue_started = None

        if self.__merengue_last_exit is None:
            self.__merengue_last_exit = time.monotonic()
        else:
            # Stop auto restart if Merengue exited less than 2 seconds ago
            if (time.monotonic() - self.__merengue_last_exit) < 2:
                self.__set_error_message(_("Workspace process error\nStopping auto restart"))
                self.__merengue_last_exit = None
                return

        self.__ui = None
        self.__set_error_message(None)

        if self.__project:
            self.__merengue.start()

    def __command_selection_changed(self, selection):
        objects = []

        for key in selection:
            obj = self.__project.get_object_by_key(key)
            objects.append(obj)

        self.__project.set_selection(objects)

    def __load_namespaces(self):
        if self.project is None:
            return

        for id, info in self.project.library_info.items():
            # Only load 3rd party libraries, Gtk ones are already loaded
            if not info.third_party or not info.enabled:
                continue

            self.__merengue_command(
                "load_namespace",
                args={
                    "namespace": info.namespace,
                    "version": info.version,
                    "object_types": info.object_types,
                },
            )

    def __set_icontheme_search_paths(self):
        if self.project is None or self.project.filename is None:
            return

        dirname = os.path.dirname(self.project.filename)

        self.__merengue_command(
            "set_icontheme_search_paths",
            args={"paths": [os.path.join(dirname, path.props.string) for path in self.project.icontheme_search_paths]}
        )

    def __on_preview_notify(self, obj, pspec):
        self.__merengue_command("set_app_property", args={"property": "preview", "value": self.preview})

    def __load_css_providers(self):
        providers = self.project.get_css_providers()

        for css in providers:
            self.__on_css_added(self.project, css)

    def __on_merengue_handle_command(self, merengue, payload):
        try:
            cmd = json.loads(payload)
            command = cmd.get("command", None)
            args = cmd.get("args", {})
        except Exception as e:
            logger.warning(f"Merengue command error: {e}")
            self.__merengue.stop()
            return

        if command == "selection_changed":
            self.__command_selection_changed(**args)
        elif command == "started":
            self.__merengue_started = True
            self.__merengue_command("gtk_settings_get", args={"property": "gtk-theme-name"})

            self.__set_icontheme_search_paths()

            self.__load_namespaces()

            self.__load_css_providers()

            self.__ui = None
            self.__on_project_selection_changed(self.__project)
        elif command == "placeholder_selected":
            self.emit(
                "placeholder-selected",
                args["ui_id"],
                args["object_id"],
                args["layout"],
                args["position"],
                args["child_type"],
            )
        elif command == "placeholder_activated":
            self.emit(
                "placeholder-activated",
                args["ui_id"],
                args["object_id"],
                args["layout"],
                args["position"],
                args["child_type"],
            )
        elif command == "gtk_settings_get":
            if args["property"] == "gtk-theme-name":
                self.__theme = args["value"]
                self.notify("gtk_theme")
        elif command == "update_ui_error":
            self.__set_error_message(args["error"])
        elif command == "css_parsing_status":
            css = self.project.get_css_by_id(args["css_id"])
            if css:
                css._emit_parsing_status(args["error_count"], args["error"], args["start"], args["end"])

    def __add_remove_placeholder(self, command, modifier):
        if self.project is None:
            return

        selection = self.project.get_selection()
        if len(selection) < 0:
            return

        obj = selection[0]
        self.__merengue_command(command, args={"ui_id": obj.ui_id, "object_id": obj.object_id, "modifier": modifier})

    def add_placeholder(self, modifier=False):
        self.__add_remove_placeholder("add_placeholder", modifier)

    def remove_placeholder(self, modifier=False):
        self.__add_remove_placeholder("remove_placeholder", modifier)


Gtk.WidgetClass.set_css_name(CmbView, "CmbView")





#!/usr/bin/pytest

"""
import .ui files into cambalache and export to compare results
"""
import os
import pytest

from cambalache import CmbProject, config


@pytest.mark.parametrize("target_tk,filename", [
    ("gtk+-3.0", "accessibility.ui"),
    ("gtk+-3.0", "bindings.ui"),
    ("gtk+-3.0", "children.ui"),
    ("gtk+-3.0", "comboboxtext.ui"),
    ("gtk+-3.0", "custom_fragment.ui"),
    ("gtk+-3.0", "custom_type.ui"),
    ("gtk+-3.0", "dialog.ui"),
    ("gtk+-3.0", "filefilter.ui"),
    ("gtk+-3.0", "label.ui"),
    ("gtk+-3.0", "levelbar.ui"),
    ("gtk+-3.0", "liststore.ui"),
    ("gtk+-3.0", "menu.ui"),
    ("gtk+-3.0", "packing.ui"),
    ("gtk+-3.0", "scale.ui"),
    ("gtk+-3.0", "signals.ui"),
    ("gtk+-3.0", "sizegroup.ui"),
    ("gtk+-3.0", "style.ui"),
    ("gtk+-3.0", "template.ui"),
    ("gtk+-3.0", "treestore.ui"),
    ("gtk+-3.0", "window.ui"),
    ("gtk-4.0", "accessibility.ui"),
    ("gtk-4.0", "bindings.ui"),
    ("gtk-4.0", "children.ui"),
    ("gtk-4.0", "comboboxtext.ui"),
    ("gtk-4.0", "custom_fragment.ui"),
    ("gtk-4.0", "custom_type.ui"),
    ("gtk-4.0", "expression.ui"),
    ("gtk-4.0", "filefilter.ui"),
    ("gtk-4.0", "inline_object.ui"),
    ("gtk-4.0", "label.ui"),
    ("gtk-4.0", "layout.ui"),
    ("gtk-4.0", "liststore.ui"),
    ("gtk-4.0", "menu.ui"),
    ("gtk-4.0", "signals.ui"),
    ("gtk-4.0", "stack_page.ui"),
    ("gtk-4.0", "string_list.ui"),
    ("gtk-4.0", "style.ui"),
    ("gtk-4.0", "template.ui"),
    ("gtk-4.0", "treestore.ui"),
    ("gtk-4.0", "ui_comments.ui"),
    ("gtk-4.0", "window.ui"),
])
def test_(target_tk, filename):
    """
    import .ui file and compare it with the exported version
    """
    path = os.path.join(os.path.dirname(__file__), target_tk, filename)
    str_original = open(path, "r").read()

    project = CmbProject(target_tk=target_tk)
    ui, msgs, detail_msg = project.import_file(path)

    assert (ui)
    assert (msgs is None)
    assert (detail_msg is None)

    str_exported = project.db.tostring(ui.ui_id)

    # Remove "Created with" comment since version will not match
    str_original = "\n".join([ l for l in str_original.splitlines() if not l.startswith(f"<!-- Created with Cambalache")])
    str_exported = "\n".join([ l for l in str_exported.splitlines() if not l.startswith(f"<!-- Created with Cambalache")])

    assert str_exported == str_original




#
# Cambalache controller object
#
# Copyright (C) 2021  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, CambalachePrivate
from . import utils
from merengue import getLogger

logger = getLogger(__name__)


class MrgController(GObject.Object):
    app = GObject.Property(type=GObject.GObject, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    object = GObject.Property(type=GObject.GObject, flags=GObject.ParamFlags.READWRITE)

    toplevel = GObject.Property(type=bool, default=False, flags=GObject.ParamFlags.READWRITE)

    selected = GObject.Property(type=bool, default=False, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        # Properties in this will will be ignored by set_object_property()
        self.property_ignore_list = set()

        self.__object = None
        self.ui_id = 0
        self.object_id = 0

        super().__init__(**kwargs)
        self.connect("notify::object", self.__on_object_changed)

        if self.object is not None:
            self.on_object_changed(self.object)

    def on_object_changed(self, value):
        old = self.__object

        if value:
            ui_id, object_id = utils.object_get_id(value).split(".")
            self.ui_id = int(ui_id)
            self.object_id = int(object_id)
        else:
            self.ui_id = 0
            self.object_id = 0

        self.__object = value

        self.object_changed(old, value)

    def __on_object_changed(self, obj, pspec):
        self.on_object_changed(self.object)

    def object_changed(self, old, new):
        pass

    # Object set property wrapper
    def set_object_property(self, name, value):
        if self.object and name not in self.property_ignore_list:
            if type(value) is str:
                CambalachePrivate.object_set_property_from_string(self.object, name, value)
            else:
                self.object.set_property(name, value)



#
# CmbFragmentEditor - Cambalache CSS Editor
#
# Copyright (C) 2022-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk
from .cmb_object import CmbObject


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_fragment_editor.ui")
class CmbFragmentEditor(Gtk.Box):
    __gtype_name__ = "CmbFragmentEditor"

    view = Gtk.Template.Child()
    child_view = Gtk.Template.Child()

    def __init__(self, **kwargs):
        self._object = None
        self.__bindings = []

        super().__init__(**kwargs)

    @GObject.Property(type=GObject.Object)
    def object(self):
        return self._object

    @object.setter
    def _set_object(self, obj):
        if obj == self._object:
            return

        for binding in self.__bindings:
            binding.unbind()

        self.__bindings = []

        self._object = obj

        if obj is None:
            return

        binding = GObject.Object.bind_property(
            obj,
            "custom-fragment",
            self.view,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        self.__bindings.append(binding)

        # Only objects have child fragments
        if type(obj) is CmbObject and obj.parent:
            binding = GObject.Object.bind_property(
                obj,
                "custom-child-fragment",
                self.child_view,
                "text",
                GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
            )
            self.__bindings.append(binding)

            self.child_view.set_visible(True)
        else:
            self.child_view.set_visible(False)


Gtk.WidgetClass.set_css_name(CmbFragmentEditor, "CmbFragmentEditor")



import ast
import sys
from lxml import etree


if __name__ == "__main__":
    filenames = sys.argv[1:]
    translatables = set()

    for filename in filenames:
        if filename.endswith('.ui'):
            root = etree.parse(filename).getroot()
            translatable_attrs = root.findall('.//*[@translatable="yes"]')
            if translatable_attrs:
                translatables.add(filename)
        elif filename.endswith('.py'):
            source = open(filename, 'r').read()
            for node in ast.iter_child_nodes(ast.parse(source)):
                if isinstance(node, ast.ImportFrom) or isinstance(node, ast.Import):
                    for alias in node.names:
                        if alias.name in ['_', 'N_']:
                            translatables.add(filename)
                            continue
        else:
            translatables.add(filename)

    for file in sorted(list(translatables)):
        print(file.removeprefix('../'))



#
# Cambalache Application
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os
import sys

from gi.repository import GLib, Gdk, Gtk, Gio, Adw, CambalachePrivate

from .cmb_window import CmbWindow
from .cmb_help_window import CmbHelpWindow
from cambalache import utils, config, _, getLogger


basedir = os.path.dirname(__file__) or "."
logger = getLogger(__name__)


class CmbApplication(Adw.Application):
    def __init__(self):
        flags = Gio.ApplicationFlags.HANDLES_OPEN

        if "CMB_APPLICATION_NON_UNIQUE" in os.environ:
            flags |= Gio.ApplicationFlags.NON_UNIQUE

        super().__init__(application_id="ar.xjuan.Cambalache", flags=flags)

        self.add_main_option("version", b"v", GLib.OptionFlags.NONE, GLib.OptionArg.NONE, _("Print version"), None)

        self.add_main_option(
            "export-all", b"E", GLib.OptionFlags.NONE, GLib.OptionArg.FILENAME, _("Deprecated: Export project"), None
        )

        self.__help_window = None
        self.__dmabuf_dialog = None

        # Dmabuf plane fd support was fixed in 4.23
        # FIXME: remove this when we depend on a fixed gtk version
        if Gtk.MINOR_VERSION <= 22:
            GLib.log_set_writer_func(self.__log_writer_handler)

    def __log_writer_handler(self, level, field_list, data):

        fields = {f.key: CambalachePrivate.log_field_get_string(f) for f in field_list if f.length < 0}

        if fields.get("GLIB_DOMAIN") == "Gdk" and fields.get("MESSAGE") == "dmabuf plane 1 has no file descriptor":
            message = "There is a bug in Gtk < 4.23 dmabuf handling, the application will crash!"
            workaround = "Run with GSK_RENDERER=gl or CASILDA_FORCE_SOFTWARE=1 as a workaround."

            if self.__dmabuf_dialog is None:
                self.__dmabuf_dialog = Gtk.AlertDialog(
                    message="Vulkan renderer bug detected",
                    detail=f"{message}\n{workaround}"
                )
                self.__dmabuf_dialog.show(self.props.active_window)

            logger.warning(message)
            logger.warning(message)

        return GLib.log_writer_default(level, field_list, data)

    def add_new_window(self):
        window = CmbWindow(application=self)
        window.connect("close-request", self.__on_window_close_request)
        window.connect("project-closed", self.__on_window_project_closed)
        self.add_window(window)
        return window

    def open_project(self, path, target_tk=None):
        window = None

        for win in self.get_windows():
            if win.project and win.project.filename == path:
                window = win

        if window is None:
            window = self.add_new_window()
            if path is not None:
                window.open_project(path, target_tk=target_tk)

        window.present()

    def import_file(self, path):
        window = self.add_new_window() if self.props.active_window is None else self.props.active_window
        window.import_file(path)
        window.present()

    def check_can_quit(self, window=None):
        windows = self.__get_windows() if window is None else [window]
        unsaved_windows = []
        windows2save = []

        def do_quit():
            if window is None:
                self.quit()
            else:
                self.remove_window(window)
                window.destroy()

        # Gather projects that needs saving
        for win in windows:
            if win.project is None:
                continue

            if win.actions["save"].get_enabled():
                unsaved_windows.append(win)

        unsaved_windows_len = len(unsaved_windows)
        if unsaved_windows_len == 0:
            do_quit()
            return

        # Create Dialog
        window = windows[0]
        dialog = window._close_project_dialog_new()

        if unsaved_windows_len > 1 or unsaved_windows[0].project.filename is None:
            # Add checkbox for each unsaved project
            box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=4)
            box.append(Gtk.Label(label=_("Select which files:"), halign=Gtk.Align.START))

            home = GLib.get_home_dir()
            untitled = 0

            for win in unsaved_windows:
                if win.project.filename is None:
                    untitled += 1

                    # Find Unique name
                    while os.path.exists(f"Untitled {untitled}.cmb"):
                        untitled += 1

                    check = Gtk.CheckButton(active=True, margin_start=8, can_focus=False)
                    entry = Gtk.Entry(text=f"Untitled {untitled}")

                    hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=4)
                    hbox.append(check)
                    hbox.append(entry)

                    box.append(hbox)
                else:
                    path = win.project.filename.replace(home, "~")
                    check = Gtk.CheckButton(label=path, active=True, margin_start=8, can_focus=False)
                    box.append(check)

                windows2save.append((win, check, entry))

            box.show()
            dialog.props.message_area.append(box)
        else:
            windows2save.append((unsaved_windows[0], None, None))

        def callback(dialog, response, window):
            dialog.destroy()

            if response == Gtk.ResponseType.ACCEPT:
                for win, check, entry in windows2save:
                    if entry is not None:
                        win.project.filename = entry.props.text
                    if check is None or check.props.active:
                        win.save_project()
            elif response == Gtk.ResponseType.CANCEL:
                return

            do_quit()

        dialog.connect("response", callback, window)
        dialog.present()

    def __get_windows(self):
        retval = []

        for win in self.get_windows():
            if isinstance(win, CmbWindow) and win.props.application is not None:
                retval.append(win)

        return retval

    def __on_window_close_request(self, window):
        self.check_can_quit(window)
        return True

    def __on_window_project_closed(self, window):
        windows = self.__get_windows()

        if len(windows) > 1:
            self.remove_window(window)
            window.destroy()

    # Action handlers
    def _on_quit_activate(self, action, data):
        self.check_can_quit()

    def _on_open_activate(self, action, data):
        filename, target_tk = data.unpack()

        # FIXME: use nullable parameter
        target_tk = target_tk if target_tk else None
        filename = filename if filename else None

        window = self.props.active_window

        if window and window.project is None:
            window.open_project(filename, target_tk)
        else:
            self.open_project(filename, target_tk)

    def _on_new_activate(self, action, data):
        target_tk, filename, uipath = data.unpack()

        # FIXME: use nullable parameter
        target_tk = target_tk if target_tk else None
        filename = filename if filename else None
        uipath = uipath if uipath else None

        window = self.props.active_window

        if window is None or window.project is not None:
            window = self.add_new_window()

        window.create_project(target_tk, filename, uipath)
        window.present()

    def _on_help_activate(self, action, data):
        if self.__help_window is None:
            self.__help_window = CmbHelpWindow(application=self)

        self.__help_window.present()

    # GApplication interface
    def do_open(self, files, nfiles, hint):
        for file in files:
            path = file.get_path()
            content_type = utils.content_type_guess(path)

            if content_type == "application/x-cambalache-project":
                self.open_project(path)
            elif content_type in ["application/x-gtk-builder", "application/x-glade"]:
                self.import_file(path)

    def do_startup(self):
        Adw.Application.do_startup(self)

        for action, accelerators, parameter_type in [
            ("quit", ["<Primary>q"], None),
            ("open", None, "(ss)"),
            ("new", None, "(sss)"),
            ("help", ["F1"], None),
        ]:
            gaction = Gio.SimpleAction.new(action, GLib.VariantType.new(parameter_type) if parameter_type else None)
            gaction.connect("activate", getattr(self, f"_on_{action}_activate"))
            self.add_action(gaction)
            if accelerators:
                self.set_accels_for_action(f"app.{action}", accelerators)

        provider = Gtk.CssProvider()
        provider.load_from_resource("/ar/xjuan/Cambalache/app/cambalache.css")
        Gtk.StyleContext.add_provider_for_display(Gdk.Display.get_default(), provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    def do_activate(self):
        if self.props.active_window is None:
            self.open_project(None)

    def do_window_removed(self, window):
        windows = self.__get_windows()

        if len(windows) == 0:
            self.activate_action("quit")

    def do_handle_local_options(self, options):
        if options.contains("version"):
            print(config.VERSION)
            return 0

        if options.contains("export-all"):
            print("Export has been deprecated and does nothing. Every UI file is updated on project save.")
            return 0

        return -1


if __name__ == "__main__":
    app = CmbApplication()
    app.run(sys.argv)



#
# CmbWindow
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os
import locale
import tempfile

from datetime import date
from gi.repository import GLib, GObject, Gio, Gdk, Gtk, Pango, Adw, GtkSource
from .cmb_tutor import CmbTutor, CmbTutorState
from .cmb_np_dialog import CmbNewProjectDialog
from .cmb_donate_dialog import CmbDonateDialog
from . import cmb_tutorial

from cambalache import (
    CmbBaseFileMonitor,
    CmbCSS,
    CmbGResource,
    CmbGResourceEditor,
    CmbObject,
    CmbProject,
    CmbProjectSettings,
    CmbTypeChooserPopover,
    CmbUI,
    _,
    config,
    getLogger,
    ngettext,
    notification_center,
    utils,
)

from cambalache.cmb_blueprint import CmbBlueprintError

logger = getLogger(__name__)

GObject.type_ensure(CmbGResourceEditor.__gtype__)


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/app/cmb_window.ui")
class CmbWindow(Adw.ApplicationWindow):
    __gtype_name__ = "CmbWindow"

    __gsignals__ = {
        "project-closed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "project-saved": (GObject.SignalFlags.RUN_FIRST, None, (CmbProject,)),
    }

    open_filter = Gtk.Template.Child()
    gtk4_filter = Gtk.Template.Child()
    gtk3_filter = Gtk.Template.Child()
    gtk_builder_filter = Gtk.Template.Child()
    blueprint_filter = Gtk.Template.Child()
    glade_filter = Gtk.Template.Child()
    css_filter = Gtk.Template.Child()
    gresource_filter = Gtk.Template.Child()

    headerbar = Gtk.Template.Child()
    title = Gtk.Template.Child()
    recent_menu = Gtk.Template.Child()
    open_button = Gtk.Template.Child()
    undo_button = Gtk.Template.Child()
    redo_button = Gtk.Template.Child()
    stack = Gtk.Template.Child()

    # Start screen
    logo = Gtk.Template.Child()
    version_label = Gtk.Template.Child()
    front_notification_list_view = Gtk.Template.Child()

    # Notification
    notification_button = Gtk.Template.Child()
    notification_list_view = Gtk.Template.Child()

    # Window message
    message_revealer = Gtk.Template.Child()
    message_label = Gtk.Template.Child()

    # Workspace
    workspace_stack = Gtk.Template.Child()
    view = Gtk.Template.Child()
    source_view = Gtk.Template.Child()
    list_view = Gtk.Template.Child()
    type_chooser = Gtk.Template.Child()
    editor_stack = Gtk.Template.Child()
    ui_editor = Gtk.Template.Child()
    object_editor = Gtk.Template.Child()
    css_editor = Gtk.Template.Child()
    gresource_editor = Gtk.Template.Child()

    # Tutor widgets
    intro_button = Gtk.Template.Child()
    menu_button = Gtk.Template.Child()

    # Properties
    current_file_object = GObject.Property(type=CmbBaseFileMonitor, flags=GObject.ParamFlags.READWRITE)
    source_style = GObject.Property(type=GtkSource.StyleScheme, flags=GObject.ParamFlags.READWRITE)

    # Settings
    completed_intro = GObject.Property(type=bool, default=False, flags=GObject.ParamFlags.READWRITE)

    MAXIMIZED = 1 << 2
    FULLSCREEN = 1 << 4

    __portal_access_msg = _(
        "Cambalache will not have access to the file for save because it is outside of your home directory. "
        "Consider using Flatseal or flatpak to give permissions to the host directory."
    )

    def __init__(self, **kwargs):
        self.__project = None
        self.__last_saved_index = None
        self.__last_saved_index_version = None

        super().__init__(**kwargs)

        # Logo animation
        def toggle_animation(gesture, n_press, x, y):
            self.logo.props.paintable.props.state = 0 if self.logo.props.paintable.props.state else 2

        click = Gtk.GestureClick()
        click.connect("released", toggle_animation)
        self.logo.add_controller(click)

        self.gtk4_import_filters = Gio.ListStore()

        for filter in [
            self.gtk4_filter,
            self.gtk_builder_filter,
            self.blueprint_filter,
            self.css_filter,
            self.gresource_filter
        ]:
            self.gtk4_import_filters.append(filter)

        self.gtk3_import_filters = Gio.ListStore()
        for filter in [self.gtk3_filter, self.gtk_builder_filter, self.glade_filter, self.css_filter, self.gresource_filter]:
            self.gtk3_import_filters.append(filter)

        self.__recent_manager = self.__get_recent_manager()

        self.actions = {}

        for action in [
            "about",
            "add_css",
            "add_gresource",
            "add_object",
            "add_object_toplevel",
            "add_placeholder",
            "add_placeholder_row",
            "add_ui",
            "clear",
            "close",
            "contact",
            "copy",
            "create_new",
            "cut",
            "debug",
            "delete",
            "donate",
            "import",
            "import_directory",
            "inspect",
            "intro",
            "liberapay",
            "open",
            "open_inspector",
            "paste",
            "patreon",
            "redo",
            "remove_parent",
            "remove_placeholder",
            "remove_placeholder_row",
            "save",
            "save_as",
            "settings",
            "show_workspace",
            "undo",
            "workspace_restart",
        ]:
            gaction = Gio.SimpleAction.new(action, None)
            gaction.connect("activate", getattr(self, f"_on_{action}_activate"))
            self.actions[action] = gaction
            self.add_action(gaction)

        # Actions with parameters and state
        for action, parameter_type, state_type, state in [
            ("add_parent", "s", None, None),
            ("open_recent", "s", None, None),
            ("workspace_theme", "s", "s", ""),
        ]:
            if state_type is None:
                gaction = Gio.SimpleAction.new(action, GLib.VariantType.new(parameter_type))
            else:
                gaction = Gio.SimpleAction.new_stateful(
                    action, GLib.VariantType.new(parameter_type), GLib.Variant(state_type, state)
                )
            gaction.connect("activate", getattr(self, f"_on_{action}_activate"))
            self.actions[action] = gaction
            self.add_action(gaction)

        # Add global accelerators
        action_map = [
            ("win.save", ["<Primary>s"]),
            ("win.save_as", ["<Shift><Primary>s"]),
            ("win.import", ["<Primary>i"]),
            ("win.close", ["<Primary>w"]),
            ("win.undo", ["<Primary>z"]),
            ("win.redo", ["<Primary><shift>z"]),
            ("win.copy", ["<Primary>c"]),
            ("win.paste", ["<Primary>v"]),
            ("win.cut", ["<Primary>x"]),
            ("win.delete", ["Delete"]),
            ("win.create_new", ["<Primary>n"]),
            ("win.open", ["<Primary>o"]),
            ("win.add_placeholder", ["<Primary>Insert", "<Primary>KP_Insert", "<Primary>KP_Add", "<Primary>plus"]),
            (
                "win.remove_placeholder",
                ["<Primary>Delete", "<Primary>KP_Delete", "<Primary>KP_Subtract", "<Primary>minus"],
            ),
            (
                "win.add_placeholder_row",
                [
                    "<Primary><shift>Insert",
                    "<Primary><shift>KP_Insert",
                    "<Primary><shift>KP_Add",
                    "<Primary><shift>plus",
                ],
            ),
            (
                "win.remove_placeholder_row",
                [
                    "<Primary><shift>Delete",
                    "<Primary><shift>KP_Delete",
                    "<Primary><shift>KP_Subtract",
                    "<Primary><shift>minus",
                ],
            ),
            ("win.show-help-overlay", ["<Primary>question"]),
            ("win.debug", ["<Shift><Primary>d"]),
        ]

        app = self.props.application
        for action, accelerators in action_map:
            app.set_accels_for_action(action, accelerators)

        # Set shortcuts window
        builder = Gtk.Builder()
        builder.add_from_resource("/ar/xjuan/Cambalache/app/cmb_shortcuts.ui")
        self.shortcut_window = builder.get_object("shortcuts")
        self.set_help_overlay(self.shortcut_window)

        # New project dialog
        self.new_project_dialog = None

        self.version_label.props.label = f"version {config.VERSION}"

        self.tutor = None
        self.tutor_waiting_for_user_action = False

        self.__clipboard_enabled = True
        self.__message_timeout_id = None

        # Create settings object
        self.settings = Gio.Settings(schema_id="ar.xjuan.Cambalache")
        self.window_settings = Gio.Settings(schema_id="ar.xjuan.Cambalache.state.window")

        # Settings list
        settings = ["completed-intro"]

        # Bind settings
        for prop in settings:
            self.settings.bind(prop, self, prop.replace("-", "_"), Gio.SettingsBindFlags.DEFAULT)

        # Force minimum fize
        self.set_default_size(320, 240)
        self.__load_window_state()
        self.__update_actions()

        app.props.style_manager.connect("notify::dark", lambda o, p: self.__update_dark_mode())

        # Bind preview
        hide_placeholders_button = Gtk.ToggleButton(tooltip_text=_("Hide placeholders"), icon_name="view-conceal-symbolic")
        self.type_chooser.content.append(hide_placeholders_button)

        GObject.Object.bind_property(
            hide_placeholders_button,
            "active",
            self.view,
            "preview",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )

        self.view.connect("notify::gtk-theme", self.__on_view_gtk_theme_notify)
        self.connect("notify::focus-widget", self.__on_focus_widget_notify)
        self.connect("close-request", self.__on_close_request)
        self.connect("map", lambda o: self.__update_dark_mode())

        self.__recent_manager.connect("changed", lambda rm: self.__update_recent_menu())
        self.__update_recent_menu()

        self.notification_list_view.notification_center = notification_center
        self.front_notification_list_view.notification_center = notification_center
        notification_center.store.connect("items-changed", self.__on_notification_center_store_items_changed)
        self.__update_notification_status()

    def __get_recent_manager(self):
        # Load the user host recently used file
        if os.environ.get("container", None) == "flatpak":
            recently_used = os.path.join(GLib.get_home_dir(), ".local", "share", "recently-used.xbel")
            if os.path.exists(recently_used):
                return Gtk.RecentManager(filename=recently_used)

        # Fallback to default
        return Gtk.RecentManager.get_default()

    def __on_view_gtk_theme_notify(self, obj, pspec):
        self.actions["workspace_theme"].set_state(GLib.Variant("s", obj.props.gtk_theme))

    @GObject.Property(type=CmbProject)
    def project(self):
        return self.__project

    @project.setter
    def _set_project(self, project):
        if self.__project:
            self.__project.disconnect_by_func(self.__on_project_filename_notify)
            self.__project.disconnect_by_func(self.__on_project_selection_changed)
            self.__project.disconnect_by_func(self.__on_project_gresource_changed)
            self.__project.disconnect_by_func(self.__on_project_changed)

        self.__project = project
        self.view.project = project
        self.type_chooser.project = project
        self.list_view.project = project

        # Clear Editors
        self.ui_editor.object = None
        self.object_editor.object = None

        if project:
            self.__project.connect("notify::filename", self.__on_project_filename_notify)
            self.__project.connect("selection-changed", self.__on_project_selection_changed)
            self.__project.connect("gresource-changed", self.__on_project_gresource_changed)
            self.__project.connect("changed", self.__on_project_changed)
            self.__on_project_selection_changed(project)

        self.__update_window_title()
        self.__update_actions()

    def __on_project_filename_notify(self, obj, pspec):
        self.__update_window_title()

    def __update_window_title(self):
        if self.project is None:
            self.title.props.title = "Cambalache"
            self.title.props.subtitle = ""
            return

        if self.project.filename:
            path = utils.friendly_homedir(self.project.filename)
        else:
            path = _("Untitled")

        prefix = "*" if self.__needs_saving() else ""

        self.title.props.title = prefix + os.path.basename(path)
        self.title.props.subtitle = os.path.dirname(path)

    @Gtk.Template.Callback("on_type_chooser_type_selected")
    def __on_type_chooser_type_selected(self, popover, info):
        selection = self.project.get_selection()
        obj = selection[0] if len(selection) else None

        if obj and not isinstance(obj, CmbObject) and not isinstance(obj, CmbUI):
            return

        device = self.get_display().get_default_seat().get_keyboard()

        # If alt is pressed, force adding object to selection
        if device and bool(device.props.modifier_state & Gdk.ModifierType.ALT_MASK):
            if obj:
                parent_id = obj.object_id if isinstance(obj, CmbObject) else None
                self.project.add_object(obj.ui_id, info.type_id, None, parent_id)
                return

        # Windows and non widgets do not need a parent
        if info.is_a("GtkWidget") and not info.is_a("GtkWindow") and info.category != "toplevel":
            # Select type and let user choose which placeholder to use
            self.type_chooser.props.selected_type = info
            self.__update_action_add_object()
        elif obj:
            # Create toplevel object/window
            self.project.add_object(obj.ui_id, info.type_id)

    @Gtk.Template.Callback("on_type_chooser_chooser_popup")
    def __on_type_chooser_chooser_popup(self, chooser, popup):
        self._show_message(_("Hold <alt> to create object in place"))

    @Gtk.Template.Callback("on_type_chooser_chooser_popdown")
    def __on_type_chooser_chooser_popdown(self, chooser, popup):
        self._show_message(None)

    @Gtk.Template.Callback("on_view_placeholder_selected")
    def __on_view_placeholder_selected(self, view, ui_id, object_id, layout, position, child_type):
        info = self.type_chooser.selected_type

        if info is not None:
            self.project.add_object(ui_id, info.type_id, None, object_id, layout, position, child_type)

        self.type_chooser.selected_type = None

    @Gtk.Template.Callback("on_view_placeholder_activated")
    def __on_view_placeholder_activated(self, view, ui_id, object_id, layout, position, child_type):
        obj = self.project.get_object_by_id(ui_id, object_id)
        popover = CmbTypeChooserPopover(pointing_to=utils.get_pointing_to(self.view), parent_type_id=obj.type_id)
        popover.set_parent(self.view)

        popover.project = self.project

        popover.connect(
            "type-selected",
            lambda o, info: self.project.add_object(ui_id, info.type_id, None, object_id, layout, position, child_type),
        )
        popover.popup()

    def __on_focus_widget_notify(self, obj, pspec):
        widget = self.props.focus_widget

        types = [Gtk.Text, Gtk.TextView]
        focused_widget_needs = True

        for type in types:
            if isinstance(widget, type):
                focused_widget_needs = False
                break

        self.__clipboard_enabled = focused_widget_needs
        self.__update_action_clipboard()

    def __update_dark_mode(self):
        if self.props.application.props.style_manager.props.dark:
            self.source_style = GtkSource.StyleSchemeManager.get_default().get_scheme("Adwaita-dark")
            paintable = Gtk.Svg(resource="/ar/xjuan/Cambalache/app/images/logo-dark.gpa")
            self.add_css_class("dark")
        else:
            paintable = Gtk.Svg(resource="/ar/xjuan/Cambalache/app/images/logo.gpa")
            self.remove_css_class("dark")
            self.source_style = GtkSource.StyleSchemeManager.get_default().get_scheme("tango")

        if fc := self.get_frame_clock():
            paintable.set_frame_clock(fc)

        paintable.props.playing = True
        paintable.props.state = 2
        self.logo.props.paintable = paintable

    def __is_project_visible(self):
        page = self.stack.get_visible_child_name()
        return self.project and page == "workspace"

    def __set_page(self, page):
        self.stack.set_visible_child_name(page)
        self.__update_actions()

    def __update_action_undo_redo(self):
        if self.__is_project_visible():
            undo_msg, redo_msg = self.project.get_undo_redo_msg()
            self.undo_button.set_tooltip_markup(f"Undo: {undo_msg}" if undo_msg is not None else None)
            self.redo_button.set_tooltip_markup(f"Redo: {redo_msg}" if redo_msg is not None else None)

            history_index = self.project.history_index
            history_index_max = self.project.history_index_max
            self.actions["undo"].set_enabled(history_index > 0)
            self.actions["redo"].set_enabled(history_index < history_index_max)
        else:
            self.actions["undo"].set_enabled(False)
            self.actions["redo"].set_enabled(False)

    def __update_action_clipboard(self):
        has_selection = False

        if self.__clipboard_enabled and self.__is_project_visible():
            sel = self.project.get_selection()
            if sel:
                # We can delete a UI too
                self.actions["delete"].set_enabled(True)

                for obj in sel:
                    if isinstance(obj, CmbObject):
                        has_selection = True
                        break

            # FIXME: Should we enable copy for CmbUI?
            for action in ["copy", "cut"]:
                self.actions[action].set_enabled(has_selection)
        else:
            for action in ["copy", "cut", "delete"]:
                self.actions[action].set_enabled(False)

        self.__update_action_clipboard_paste()

    def __update_action_clipboard_paste(self):
        if self.__clipboard_enabled and self.__is_project_visible():
            self.actions["paste"].set_enabled(self.project.clipboard_count() > 0)
        else:
            self.actions["paste"].set_enabled(False)

    def __on_project_changed(self, project):
        self.__update_action_undo_redo()
        self.__update_action_save()

    def __update_gresource_view(self):
        if self.workspace_stack.get_visible_child_name() != "gresource":
            return

        sel = self.project.get_selection()
        if len(sel) < 1:
            return

        obj = sel[0]

        if isinstance(obj, CmbGResource):
            gresource_id = obj.gresources_bundle.gresource_id
            resource_xml = self.project.db.gresource_tostring(gresource_id)
        else:
            resource_xml = ""

        self.source_view.buffer.set_text(resource_xml)

    def __on_project_gresource_changed(self, project, gresource, field):
        self.__update_gresource_view()

    def __on_project_selection_changed(self, project):
        selection = project.get_selection()
        self.__update_action_clipboard()

        obj = selection[0] if selection else None

        if isinstance(obj, CmbUI):
            self.ui_editor.object = obj
            self.current_file_object = obj
            self.workspace_stack.set_visible_child_name("ui")
            self.editor_stack.set_visible_child_name("ui")
            obj = None
        elif isinstance(obj, CmbObject):
            self.current_file_object = obj.ui
            self.workspace_stack.set_visible_child_name("ui")
            self.editor_stack.set_visible_child_name("object")
            if obj:
                self.__user_message_by_type(obj.info)
        elif isinstance(obj, CmbCSS):
            self.current_file_object = None
            self.css_editor.object = obj
            self.editor_stack.set_visible_child_name("css")
            obj = None
        elif isinstance(obj, CmbGResource):
            self.current_file_object = obj
            self.gresource_editor.object = obj
            self.workspace_stack.set_visible_child_name("gresource")
            self.editor_stack.set_visible_child_name("gresource")
            self.__update_gresource_view()
            obj = None

        self.object_editor.object = obj

        self.__update_action_add_object()
        self.__update_action_remove_parent()

        if not selection:
            self.workspace_stack.set_visible_child_name("ui")
            self.editor_stack.set_visible_child_name("empty")

    def __update_action_intro(self):
        enabled = False

        if not self.completed_intro:
            enabled = True
            self.intro_button.props.tooltip_text = _("Start interactive introduction")

        self.intro_button.set_visible(enabled)

    def __update_action_add_object(self):
        has_project = self.__is_project_visible()
        has_selection = True if self.project and len(self.project.get_selection()) > 0 else False
        has_info = self.type_chooser.props.selected_type is not None
        enabled = has_project and has_selection and has_info

        for action in ["add_object", "add_object_toplevel"]:
            self.actions[action].set_enabled(enabled)

    def __update_action_remove_parent(self):
        if self.project is None:
            self.actions["remove_parent"].set_enabled(False)
            return

        selection = self.project.get_selection()
        obj = selection[0] if len(selection) else None

        if obj and isinstance(obj, CmbObject) and obj.parent_id and obj.parent.n_items == 1:
            self.actions["remove_parent"].set_enabled(True)
            return

        self.actions["remove_parent"].set_enabled(False)

    def __needs_saving(self):
        if self.project is None:
            return False

        return self.__is_project_visible() and (
            self.project.history_index != self.__last_saved_index or
            self.project.history_index_version != self.__last_saved_index_version
        )

    def __update_action_save(self):
        changed = self.__needs_saving()

        self.__update_window_title()

        self.actions["save"].set_enabled(changed)
        if changed:
            self.title.add_css_class("changed")
        else:
            self.title.remove_css_class("changed")

    def __update_actions(self):
        has_project = self.__is_project_visible()

        for action in [
            "add_css",
            "add_gresource",
            "add_ui",
            "close",
            "debug",
            "delete",
            "import",
            "import_directory",
            "save_as",
            "settings",
        ]:
            self.actions[action].set_enabled(has_project)

        self.__update_action_save()
        self.__update_action_intro()
        self.__update_action_clipboard()
        self.__update_action_undo_redo()
        self.__update_action_add_object()
        self.__update_action_remove_parent()

    def __file_open_dialog_new(self, title, filter_obj=None, accept_label=None, use_project_dir=False, filters=None):
        dialog = Gtk.FileDialog(
            modal=True,
            title=title,
            default_filter=filter_obj,
            filters=filters,
            accept_label=accept_label,
        )

        if use_project_dir and self.project and self.project.filename:
            dialog.set_initial_folder(Gio.File.new_for_path(os.path.dirname(self.project.filename)))

        return dialog

    def present_message_to_user(self, message, secondary_text=None, details=None):
        # TODO: replace with custom widget

        dialog = Gtk.MessageDialog(
            transient_for=self,
            text=message,
            secondary_text=secondary_text,
            modal=True,
            use_markup=True,
            secondary_use_markup=True,
        )

        dialog.add_button(_("Copy to clipboard"), 1)
        dialog.add_button(_("Close"), Gtk.ResponseType.CLOSE)

        if details:
            sw = Gtk.ScrolledWindow(
                vexpand=True,
                propagate_natural_width=True,
                propagate_natural_height=True,
                max_content_width=800,
                max_content_height=480,
            )
            box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=4)
            sw.set_child(box)

            for detail in details:
                box.append(
                    Gtk.Label(
                        label=detail,
                        use_markup=True,
                        halign=Gtk.Align.START,
                        xalign=0.0,
                        max_width_chars=120,
                        wrap_mode=Pango.WrapMode.WORD,
                    )
                )
            dialog.props.message_area.append(sw)

        def on_response(dialog, response):
            if response == 1:
                clip = self.get_clipboard()
                details_string = '\n'.join(details or [])
                clip.set(f"{message}\n{secondary_text}\n{details_string}")
            else:
                dialog.destroy()

        dialog.connect("response", on_response)
        dialog.present()

    def __check_if_filename_is_in_portal(self, filename):
        if os.environ.get("FLATPAK_ID", None) != "ar.xjuan.Cambalache":
            return False

        return filename.startswith(f"/run/user/{os.getuid()}/doc/")

    def import_ui(self, filename, target_tk=None, autoselect=True, present_errors=True):
        if self.project is None:
            dirname = os.path.dirname(filename)
            basename = os.path.basename(filename)
            name, ext = os.path.splitext(basename)

            if not target_tk:
                target_tk = CmbProject.get_target_from_ui_file(filename)

            if not target_tk:
                self.ask_gtk_version(filename)
                return (None, None)

            self.project = CmbProject(filename=os.path.join(dirname, f"{name}.cmb"), target_tk=target_tk)
            self.__set_page("workspace")
            self.__update_actions()

        # Check if its already imported
        ui = self.project.get_ui_by_filename(filename)

        if ui:
            if autoselect:
                self.project.set_selection([ui])
            return (None, None)

        try:
            if self.__check_if_filename_is_in_portal(filename):
                raise Exception(self.__portal_access_msg)

            ui, msg, detail = self.project.import_file(filename)

            if autoselect:
                self.project.set_selection([ui])

            if msg:
                details = "\n".join(detail)
                logger.warning(f"Error parsing {filename}\n{details}")

                filename = os.path.basename(filename)
                name, ext = os.path.splitext(filename)
                title = _("Error importing {filename}").format(filename=os.path.basename(filename))

                if len(msg) > 1:
                    # Translators: This is used to create a unordered list of unsupported features to show the user
                    list = [_("    • {message}").format(message=message) for message in msg]

                    # Translators: This will be the heading of a list of unsupported features
                    first_msg = _("Cambalache encounter the following issues:")

                    # Translators: this is the last message after the list of unsupported features
                    last_msg = _("Your file will be saved as '{name}.cmb.ui' to avoid data loss.").format(name=name)

                    unsupported_features_list = [first_msg] + list + [last_msg]
                    if present_errors:
                        self.present_message_to_user(title, details=unsupported_features_list)

                    return (title, "\n".join(unsupported_features_list))
                else:
                    unsupported_feature = msg[0]
                    text = _(
                        "Cambalache encounter {unsupported_feature}\n"
                        "Your file will be saved as '{name}.cmb.ui' to avoid data loss."
                    ).format(unsupported_feature=unsupported_feature, name=name)

                    if present_errors:
                        self.present_message_to_user(title, secondary_text=text)

                    return (title, text)

            # All good!
            return (None, None)
        except Exception as e:
            filename = os.path.basename(filename)
            logger.warning(f"Error loading {filename}", exc_info=True)

            title = _("Exception importing {filename}").format(filename=filename)
            msg = str(e)

            if present_errors:
                self.present_message_to_user(title, secondary_text=msg)

            return (title, msg)

    def ask_gtk_version(self, filename):
        basename = os.path.basename(filename)

        dialog = Gtk.MessageDialog(
            transient_for=self,
            message_type=Gtk.MessageType.QUESTION,
            text=_("Which is the target Gtk version of {filename}?").format(filename=basename),
            modal=True,
        )

        dialog.add_button("Gtk 3", 3)
        dialog.add_button("Gtk 4", 4)
        dialog.set_default_response(4)

        def on_ask_gtk_version_response(d, r):
            target_tk = "gtk-4.0" if r == 4 else "gtk+-3.0"
            self.import_ui(filename, target_tk=target_tk)
            d.destroy()

        dialog.connect("response", on_ask_gtk_version_response)
        dialog.present()

    def create_project(self, target_tk, filename, uipath):
        if self.project:
            return

        self.project = CmbProject(filename=filename, target_tk=target_tk)
        self.__last_saved_index = self.project.history_index
        self.__last_saved_index_version = self.project.history_index_version

        # Create UI and select it
        ui = self.project.add_ui(uipath)
        self.project.set_selection([ui])

        self.__set_page("workspace")
        self.__update_actions()

    def open_project(self, filename, target_tk):
        try:
            if filename is not None:
                content_type = utils.content_type_guess(filename)

                if content_type in ["application/x-gtk-builder", "application/x-glade"]:
                    if target_tk is None:
                        target_tk = CmbProject.get_target_from_ui_file(filename)

                    if target_tk is None and filename is not None:
                        self.ask_gtk_version(filename)
                        return

                    self.import_ui(filename, target_tk=target_tk)
                elif content_type != "application/x-cambalache-project":
                    raise Exception(_("Unknown file type {content_type}").format(content_type=content_type))

            if self.project is None:
                self.project = CmbProject(filename=filename, target_tk=target_tk)

            self.__last_saved_index = self.project.history_index
            self.__last_saved_index_version = self.project.history_index_version
            self.__set_page("workspace")
            self.__update_actions()
        except Exception as e:
            filename = os.path.basename(filename)
            logger.warning(f"Error loading {filename}", exc_info=True)
            self.present_message_to_user(
                _("Error loading {filename}").format(filename=filename), secondary_text=str(e)
            )

    def __app_activate_open(self, filename, target_tk=None):
        if self.__check_if_filename_is_in_portal(filename):
            self.present_message_to_user(
                _("Error opening {filename}").format(filename=os.path.basename(filename)),
                secondary_text=self.__portal_access_msg
            )
            return

        self.props.application.activate_action("open", GLib.Variant("(ss)", (filename, target_tk or "")))

    def _on_open_activate(self, action, data):
        def dialog_callback(dialog, res):
            try:
                file = dialog.open_finish(res)
                self.__app_activate_open(file.get_path())
            except Exception as e:
                logger.warning(f"Error {e}")

        dialog = self.__file_open_dialog_new(_("Choose project to open"), filter_obj=self.open_filter)
        dialog.open(self, None, dialog_callback)

    def _on_create_new_activate(self, action, data):
        self.new_project_dialog = CmbNewProjectDialog()
        self.new_project_dialog.connect("create-new-project", self.__on_create_new_project)
        self.new_project_dialog.present(self)

    def __on_create_new_project(self, dialog, target_tk, filename, uipath):
        # Close dialog
        self.new_project_dialog.close()
        self.new_project_dialog = None

        # Emit application new action
        self.props.application.activate_action("new", GLib.Variant("(sss)", (target_tk, filename, uipath)))

        # Switch to workspace page if project was created
        self.__set_page("workspace" if self.project else "cambalache")

    def __on_undo_redo_activate(self, undo):
        if self.project is None:
            return
        try:
            if undo:
                self.project.undo()
            else:
                self.project.redo()
        except Exception as e:
            logger.warning("Undo/Redo error", exc_info=True)
            self.present_message_to_user(
                _("Undo/Redo stack got corrupted"),
                secondary_text=_("Please try to reproduce and file an issue\n Error: {msg}").format(msg=str(e))
            )
        self.__update_action_undo_redo()

    def _on_undo_activate(self, action, data):
        self.__on_undo_redo_activate(True)

    def _on_redo_activate(self, action, data):
        self.__on_undo_redo_activate(False)

    def _on_save_activate(self, action, data):
        self.save_project()

    def __save_dialog_callback(self, dialog, res):
        filename = None
        try:
            file = dialog.save_finish(res)
            filename = file.get_path()
        except Exception:
            return

        if not filename:
            return

        try:
            if self.__check_if_filename_is_in_portal(filename):
                raise Exception(self.__portal_access_msg)

            self.project.filename = filename
            self.__save()
        except Exception as e:
            filename = os.path.basename(filename) if filename else filename
            logger.warning(f"Error saving {filename}", exc_info=True)
            self.present_message_to_user(
                _("Error importing {filename}").format(filename=filename),
                secondary_text=str(e)
            )

    def _on_save_as_activate(self, action, data):
        if self.project is None:
            return

        dialog = self.__file_open_dialog_new(_("Choose a new file to save the project"))
        dialog.save(self, None, self.__save_dialog_callback)

    def _on_add_ui_activate(self, action, data):
        if self.project is None:
            return

        ui = self.project.add_ui()
        self.project.set_selection([ui])

    def _on_add_css_activate(self, action, data):
        if self.project is None:
            return

        css = self.project.add_css()
        self.project.set_selection([css])

    def __remove_object_with_confirmation(self, obj):
        dialog = Gtk.MessageDialog(
            transient_for=self,
            modal=True,
            message_type=Gtk.MessageType.QUESTION,
            buttons=Gtk.ButtonsType.YES_NO,
            text=_("Do you really want to remove {name}?").format(name=obj.display_name),
            use_markup=True,
        )

        def on_dialog_response(dialog, response):
            if response == Gtk.ResponseType.YES:
                if isinstance(obj, CmbUI):
                    self.project.remove_ui(obj)
                elif isinstance(obj, CmbCSS):
                    self.project.remove_css(obj)
                elif isinstance(obj, CmbGResource):
                    self.project.remove_gresource(obj)

            dialog.destroy()

        dialog.connect("response", on_dialog_response)
        dialog.present()

    def _on_copy_activate(self, action, data):
        if self.project:
            self.project.copy()
            self.__update_action_clipboard_paste()

    def _on_paste_activate(self, action, data):
        if self.project:
            self.project.paste()
            self.__update_action_clipboard_paste()

    def _on_cut_activate(self, action, data):
        if self.project:
            self.project.cut()
            self.__update_action_clipboard_paste()

    def _on_delete_activate(self, action, data):
        if self.project is None:
            return

        selection = self.project.get_selection()
        for obj in selection:
            try:
                if isinstance(obj, CmbObject):
                    self.project.remove_object(obj)
                elif isinstance(obj, CmbGResource):
                    if obj.resource_type == "gresources":
                        self.__remove_object_with_confirmation(obj)
                    else:
                        self.project.remove_gresource(obj)
                else:
                    self.__remove_object_with_confirmation(obj)
            except Exception as e:
                self.present_message_to_user(
                    _("Error deleting {name}").format(name=obj.display_name),
                    secondary_text=str(e)
                )

    def _on_add_object_activate(self, action, data):
        info = self.type_chooser.props.selected_type
        if self.project is None or info is None:
            return

        selection = self.project.get_selection()
        if len(selection) > 0:
            obj = selection[0]
            parent_id = obj.object_id if isinstance(obj, CmbObject) else None
            self.project.add_object(obj.ui_id, info.type_id, None, parent_id)
            return

    def _on_add_object_toplevel_activate(self, action, data):
        info = self.type_chooser.props.selected_type
        if self.project is None or info is None:
            return

        selection = self.project.get_selection()
        if len(selection) > 0:
            obj = selection[0]
            self.project.add_object(obj.ui_id, info.type_id)
            return

    def _on_clear_activate(self, action, data):
        selection = self.project.get_selection()
        if len(selection) > 0:
            for obj in selection:
                if isinstance(obj, CmbObject):
                    obj.clear_properties()

    def __present_import_error(self, filename, msg, detail):
        details = "\n".join(detail)
        logger.warning(f"Error parsing {filename}\n{details}")

        filename = os.path.basename(filename)
        name, ext = os.path.splitext(filename)
        unsupported_features_list = None
        text = None

        if len(msg) > 1:
            # Translators: This is used to create a unordered list of unsupported features to show the user
            list = [_("    • {message}").format(message=message) for message in msg]

            # Translators: This will be the heading of a list of unsupported features
            first_msg = _("Cambalache encounter the following issues:")

            # Translators: this is the last message after the list of unsupported features
            last_msg = _("Your file will be saved as '{name}.cmb.ui' to avoid data loss.").format(name=name)

            unsupported_features_list = [first_msg] + list + [last_msg]
        else:
            unsupported_feature = msg[0]
            text = _(
                "Cambalache encounter {unsupported_feature}\nYour file will be saved as '{name}.cmb.ui' to avoid data loss."
            ).format(unsupported_feature=unsupported_feature, name=name)

        self.present_message_to_user(
            _("Error importing {filename}").format(filename=filename),
            secondary_text=text,
            details=unsupported_features_list,
        )

    def import_file(self, path, autoselect=True, present_errors=True):
        content_type = utils.content_type_guess(path)

        if content_type in ["application/x-gtk-builder", "application/x-glade", "text/x-blueprint"]:
            return self.import_ui(path, autoselect=autoselect, present_errors=present_errors)
        elif content_type == "text/css":
            self.project.add_css(path)
        elif content_type == "application/xml" and path.endswith("gresource.xml"):
            self.project.import_gresource(path)

        return (None, None)

    def _on_import_activate(self, action, data):
        if self.project is None:
            return

        def dialog_callback(dialog, res):
            try:
                for file in dialog.open_multiple_finish(res):
                    self.import_file(file.get_path())
            except Exception as e:
                logger.warning(f"Error {e}")

        if self.project.target_tk == "gtk-4.0":
            import_filters = self.gtk4_import_filters
        else:
            import_filters = self.gtk3_import_filters

        dialog = self.__file_open_dialog_new(
            _("Choose file to import"),
            filters=import_filters,
            accept_label=_("Import"),
            use_project_dir=True
        )
        dialog.open_multiple(self, None, dialog_callback)

    def _on_import_directory_activate(self, action, data):
        if self.project is None:
            return

        def progress_dialog_new():
            dialog = Gtk.Window(
                title="Importing directory",
                transient_for=self,
                default_width=640,
                modal=True
            )
            progressbar = Gtk.ProgressBar(
                text=_("Loading directory contents"),
                ellipsize=Pango.EllipsizeMode.START,
                show_text=True,
                vexpand=True,
                margin_top=32,
                margin_bottom=32,
                margin_start=16,
                margin_end=16,
            )
            dialog.set_child(progressbar)
            dialog.present()

            return dialog, progressbar

        def dialog_callback(dialog, res):
            main_loop = GLib.MainContext.default()

            try:
                dir = dialog.select_folder_finish(res)
                dirpath = dir.get_path()

                progress, progressbar = progress_dialog_new()

                def pulse():
                    progressbar.pulse()
                    while main_loop.pending():
                        main_loop.iteration(False)

                files = self.project._list_supported_files(dirpath, pulse)
                n_files = len(files)

                self.project.history_push(_('Import {n} directory "{dirpath}"').format(dirpath=dirpath, n=n_files))

                basedir = self.project.dirname + "/"

                errors = []

                for i, path in enumerate(files):
                    progressbar.set_text(path.removeprefix(basedir))
                    progressbar.set_fraction(i/n_files)
                    error, error_details = self.import_file(path, autoselect=False, present_errors=False)

                    if error:
                        errors.append(f"\n<b>{error}</b>\n{error_details}")

                    while main_loop.pending():
                        main_loop.iteration(False)

                self.project.history_pop()

                if errors:
                    text = _("{errors} files out of {n} had errors while loading").format(errors=len(errors), n=n_files)
                    self.present_message_to_user(
                        _("Error importing directory {basename}").format(basename=os.path.basename(dirpath)),
                        secondary_text=text,
                        details=errors
                    )
                else:
                    self._show_message(_("Imported {n} files").format(n=n_files))

                progress.close()

            except Exception as e:
                logger.warning(f"Error {e}")

        dialog = self.__file_open_dialog_new(
            _("Choose directory to import"),
            accept_label=_("Import directory"),
            use_project_dir=True
        )
        dialog.select_folder(self, None, dialog_callback)

    def _on_add_gresource_activate(self, action, data):
        if self.project is None:
            return

        gresource = self.project.add_gresource("gresources")
        self.project.set_selection([gresource])

    def __save(self):
        retval = False

        try :
            retval = self.project.save()
        except CmbBlueprintError as e:
            self.present_message_to_user(
                _("Error saving project"),
                secondary_text=ngettext("blueprintcompiler encounter the following error:", "blueprintcompiler encounter the following errors:", len(e.errors)),
                details=[str(e)]
            )
        finally:
            if retval:
                self.__last_saved_index = self.project.history_index
                self.__last_saved_index_version = self.project.history_index_version
                self.__update_action_save()
                self.emit("project-saved", self.project)

    def save_project(self):
        if self.project is None:
            return False

        if self.project.filename is None:
            dialog = self.__file_open_dialog_new(_("Choose a file to save the project"))
            dialog.save(self, None, self.__save_dialog_callback)
            return True

        # Save project and update last saved index
        self.__save()

        return False

    def _close_project_dialog_new(self):
        text = _("Save changes before closing?")
        dialog = Gtk.MessageDialog(
            transient_for=self,
            message_type=Gtk.MessageType.QUESTION,
            text=f"<b><big>{text}</big></b>",
            use_markup=True,
            modal=True,
        )

        # Add buttons
        dialog.add_buttons(
            _("Close without Saving"),
            Gtk.ResponseType.CLOSE,
            _("Cancel"),
            Gtk.ResponseType.CANCEL,
            _("Save"),
            Gtk.ResponseType.ACCEPT,
        )

        dialog.set_default_response(Gtk.ResponseType.ACCEPT)
        return dialog

    def _on_close_activate(self, action, data):
        def close_project():
            self.project = None
            self.__set_page("cambalache")
            self.emit("project-closed")

        if self.actions["save"].get_enabled():
            dialog = self._close_project_dialog_new()

            def callback(dialog, response, window):
                dialog.destroy()

                if response == Gtk.ResponseType.ACCEPT:
                    if self.project.filename is None:
                        def save_callback(dialog, res):
                            try:
                                file = dialog.save_finish(res)
                                self.project.filename = file.get_path()
                                self.save_project()
                                close_project()
                            except Exception as e:
                                logger.warning(f"Error {e}")

                        file_dialog = self.__file_open_dialog_new(_("Choose a new file to save the project"))
                        file_dialog.save(self, None, save_callback)
                    else:
                        self.save_project()
                        close_project()
                elif response == Gtk.ResponseType.CLOSE:
                    close_project()

            dialog.connect("response", callback, self)
            dialog.present()
        else:
            close_project()

    def _on_settings_activate(self, action, data):
        if self.project is None:
            return

        settings = CmbProjectSettings(project=self.project)
        settings.present(self)

    def _on_debug_activate(self, action, data):
        if self.project.filename:
            filename = self.project.filename + ".db"
        else:
            fd, filename = tempfile.mkstemp(".db", "cmb")

        self.project.db_move_to_fs(filename)
        Gtk.show_uri(self, f"file://{filename}", Gdk.CURRENT_TIME)

    def __populate_supporters(self, about):
        gbytes = Gio.resources_lookup_data("/ar/xjuan/Cambalache/app/SUPPORTERS.md", Gio.ResourceLookupFlags.NONE)
        supporters = gbytes.get_data().decode("UTF-8").splitlines()
        sponsors = []

        for name in supporters:
            if name.startswith(" - "):
                sponsors.append(name[3:])

        about.add_credit_section(_("Supporters"), sponsors)

    def __update_translators(self, about):
        lang_country, encoding = locale.getlocale()
        lang = lang_country.split("_")[0]

        translators = {
            "cs": ["Daniel Rusek", "Vojtěch Perník"],
            "de": ["PhilProg", "Philipp Unger"],
            "es": ["Juan Pablo Ugarte"],
            "fi": ["Erwinjitsu", "Ilmari Lauhakangas"],
            "fr": ["rene-coty"],
            "it": ["Lorenzo Capalbo"],
            "nl": ["Gert"],
            "pt_BR": ["John peter sa"],
            "ro": ["Vlad"],
            "sv": ["Anders Jonsson"],
            "uk": ["Volodymyr M. Lisivka"],
            "zh_CN": ["xu-haozhe"],
        }

        translator_list = translators.get(lang, None)

        if translator_list:
            about.props.translator_credits = "\n".join(translator_list)

    def _on_about_activate(self, action, data):
        about = Adw.AboutDialog.new_from_appdata("/ar/xjuan/Cambalache/app/metainfo.xml", config.VERSION)

        about.props.artists = [
            "Franco Dodorico",
            "Juan Pablo Ugarte",
        ]
        about.props.copyright = f"© 2020-{date.today().year} Juan Pablo Ugarte"
        about.props.license_type = Gtk.License.LGPL_2_1_ONLY

        self.__update_translators(about)
        self.__populate_supporters(about)

        about.present(self)

    def _on_add_parent_activate(self, action, data):
        obj = self.project.get_selection()[0]
        self.project.add_parent(data.get_string(), obj)

    def _on_donate_activate(self, action, data):
        dialog = CmbDonateDialog()
        dialog.present(self)

    def _on_liberapay_activate(self, action, data):
        Gtk.show_uri(self, "https://liberapay.com/xjuan/donate", Gdk.CURRENT_TIME)

    def _on_patreon_activate(self, action, data):
        Gtk.show_uri(self, "https://www.patreon.com/cambalache", Gdk.CURRENT_TIME)

    def _on_contact_activate(self, action, data):
        Gtk.show_uri(self, "https://matrix.to/#/#cambalache:gnome.org", Gdk.CURRENT_TIME)

    def _on_add_placeholder_activate(self, action, data):
        self.view.add_placeholder()

    def _on_remove_placeholder_activate(self, action, data):
        self.view.remove_placeholder()

    def _on_add_placeholder_row_activate(self, action, data):
        self.view.add_placeholder(modifier=True)

    def _on_remove_placeholder_row_activate(self, action, data):
        self.view.remove_placeholder(modifier=True)

    def _on_remove_parent_activate(self, action, data):
        selection = self.project.get_selection()
        if selection is None or len(selection) == 0:
            return

        self.project.remove_parent(selection[0])

    def _on_show_workspace_activate(self, action, data):
        self.__set_page("workspace" if self.project else "cambalache")

    def __clear_tutor(self):
        try:
            self.disconnect_by_func(self.__on_project_notify)
            self.project.disconnect_by_func(self.__on_ui_added)
            self.project.disconnect_by_func(self.__on_object_added)
        except Exception:
            pass
        self.tutor = None

    def __on_project_notify(self, obj, pspec):
        if self.project:
            self.tutor_waiting_for_user_action = False
            self.tutor.play()
            self.disconnect_by_func(self.__on_project_notify)

    def __on_object_added(self, project, obj, data):
        if obj.info.is_a(data):
            project.disconnect_by_func(self.__on_object_added)
            self.tutor_waiting_for_user_action = False
            self.tutor.play()

    def __on_ui_added(self, project, ui):
        self.tutor_waiting_for_user_action = False
        project.disconnect_by_func(self.__on_ui_added)
        self.tutor.play()

    def __on_tutor_show_node(self, tutor, node, widget):
        if node == "add-project":
            if self.project is None:
                self.connect("notify::project", self.__on_project_notify)
        elif node == "add-ui":
            self.project.connect("ui-added", self.__on_ui_added)
        elif node == "add-window":
            self.project.connect("object-added", self.__on_object_added, "GtkWindow")
        elif node == "add-grid":
            self.project.connect("object-added", self.__on_object_added, "GtkGrid")
        elif node == "add-button":
            self.project.connect("object-added", self.__on_object_added, "GtkButton")
        elif node == "show-type-popover":
            widget.props.popover.popup()
        elif node == "show-type-popover-gtk":
            child = utils.widget_get_children(widget)[0]
            child.props.popover.popup()

    def __on_tutor_hide_node(self, tutor, node, widget):
        if node == "intro-end":
            self.completed_intro = True
            self.__clear_tutor()
        elif node == "add-project":
            if self.__project is None:
                self.tutor_waiting_for_user_action = True
                self.tutor.pause()
        elif node in ["add-ui", "add-window", "add-grid", "add-button"]:
            self.tutor_waiting_for_user_action = True
            self.tutor.pause()
        elif node in ["donate"]:
            self.menu_button.popdown()
        elif node == "show-type-popover":
            widget.props.popover.popdown()
        elif node == "show-type-popover-gtk":
            child = utils.widget_get_children(widget)[0]
            child.props.popover.popdown()

        self.__update_actions()

    def _on_intro_activate(self, action, data):
        if self.tutor_waiting_for_user_action:
            return

        if self.tutor:
            if self.tutor.state == CmbTutorState.PLAYING:
                self.tutor.pause()
            else:
                self.tutor.play()
            return

        # Ensure button is visible and reset config flag since we are playing
        # the tutorial from start
        self.intro_button.set_visible(True)
        self.completed_intro = False

        self.tutor = CmbTutor(script=cmb_tutorial.intro, window=self)
        self.tutor.connect("show-node", self.__on_tutor_show_node)
        self.tutor.connect("hide-node", self.__on_tutor_hide_node)
        self.tutor.play()

    def _on_workspace_restart_activate(self, action, data):
        self.view.restart_workspace()

    def _on_workspace_theme_activate(self, action, data):
        self.view.props.gtk_theme = data.get_string()
        action.set_state(data)

    def _on_inspect_activate(self, action, data):
        self.view.inspect()

    def _on_open_inspector_activate(self, action, data):
        self.view.set_interactive_debugging(True)

    def _on_open_recent_activate(self, action, data):
        self.__app_activate_open(data.get_string(), "")

    def __update_recent_menu(self):
        mime_types = ["application/x-cambalache-project"]

        self.recent_menu.remove_all()

        for recent in self.__recent_manager.get_items():
            if recent.get_mime_type() not in mime_types:
                continue

            if not recent.exists() or recent.get_age() > 7:
                continue

            filename, host = GLib.filename_from_uri(recent.get_uri())

            item = Gio.MenuItem()
            item.set_label(recent.get_display_name().replace("_", "__"))
            item.set_action_and_target_value("win.open_recent", GLib.Variant("s", filename))
            self.recent_menu.append_item(item)

        # set menu if there is anything
        self.open_button.props.menu_model = self.recent_menu if self.recent_menu.get_n_items() else None

    def __load_window_state(self):
        state = self.window_settings.get_uint("state")

        if state & self.MAXIMIZED:
            self.maximize()
        elif state & self.FULLSCREEN:
            self.fullscreen()
        else:
            w, h = self.window_settings.get_value("size").unpack()
            if w and h:
                self.set_default_size(w, h)

    def __save_window_state(self):
        fullscreen = self.props.fullscreened
        maximized = self.props.maximized
        state = 0

        if fullscreen:
            state = state | self.FULLSCREEN

        if maximized:
            state = state | self.MAXIMIZED

        # Maintain compatibility with Gtk 3 state
        self.window_settings.set_uint("state", state)

        size = (0, 0) if fullscreen or maximized else (self.props.default_width, self.props.default_height)

        self.window_settings.set_value("size", GLib.Variant("(ii)", size))

    def __on_close_request(self, window):
        self.__save_window_state()
        return False

    def __user_message_by_type(self, info):
        msg = None

        # TODO: Move this strings to the database, so it can be defined in 3rd party plugins too
        if info.is_a("GtkBox"):
            msg = _("<Ctrl>+Ins/Del to add/remove placeholders")
        elif info.is_a("GtkGrid"):
            msg = _("<Ctrl>+Ins/Del to add/remove columns\n<Shift>+<Ctrl>+Ins/Del to add/remove rows")
        elif info.is_a("GtkAssistant") or info.is_a("GtkStack"):
            msg = _("<Ctrl>+Ins/Del to add/remove pages")

        self._show_message(msg)

    def __on_message_timeout(self, data):
        self.__message_timeout_id = None
        self.message_revealer.props.reveal_child = False
        return GLib.SOURCE_REMOVE

    def _show_message(self, msg):
        if self.__message_timeout_id:
            GLib.source_remove(self.__message_timeout_id)
            self.__message_timeout_id = None

        if msg:
            self.message_label.props.label = msg
            self.message_revealer.props.reveal_child = True
            self.__message_timeout_id = GLib.timeout_add(len(msg) * 100, self.__on_message_timeout, None)
        else:
            self.message_revealer.props.reveal_child = False

    @Gtk.Template.Callback("on_notification_close_button")
    def __on_notification_close_button(self, button):
        self.notification_button.props.popover.props.autohide = True
        self.notification_button.popdown()

    def __on_notification_popover_timeout(self, data):
        self.notification_button.props.popover.props.autohide = True
        self.notification_button.popdown()
        return GLib.SOURCE_REMOVE

    def __on_notification_center_store_items_changed(self, store, position, removed, added):
        self.__update_notification_status()

        if added and self.stack.get_visible_child_name() != "cambalache":
            # Set autohide to false, otherwise popover is immediately closed and input does not work anymore
            self.notification_button.props.popover.props.autohide = False
            self.notification_button.popup()
            GLib.timeout_add_seconds(60, self.__on_notification_popover_timeout, None)

    def __update_notification_status(self):
        enabled = bool(notification_center.store.props.n_items > 0)

        if enabled:
            self.notification_button.remove_css_class("hidden")
            self.notification_button.set_visible(True)
            self.notification_button.set_sensitive(True)
        else:
            self.notification_button.add_css_class("hidden")
            self.notification_button.set_sensitive(False)
            self.notification_button.popdown()





# GtkStack Controller
#
# Copyright (C) 2022  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .mrg_gtk_widget import MrgGtkWidget
from merengue import MrgPlaceholder


class MrgGtkStack(MrgGtkWidget):
    object = GObject.Property(type=Gtk.Stack, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __get_placeholder(self):
        for child in self.get_children():
            if isinstance(child, MrgPlaceholder):
                return child
        return None

    def __ensure_placeholders(self):
        if self.object is None:
            return

        if len(self.get_children()) == 0:
            self.add(MrgPlaceholder(visible=True, controller=self))

    def object_changed(self, old, new):
        super().object_changed(old, new)
        self.__ensure_placeholders()

    def show_child(self, child):
        self.object.set_visible_child(child)

    def add(self, child):
        if self.object is None:
            return

        if Gtk.MAJOR_VERSION == 3:
            self.object.add(child)
        else:
            self.object.add_child(child)

    def remove_child(self, child):
        if self.object is None:
            return

        if Gtk.MAJOR_VERSION == 4:
            self.object.remove(child)
        else:
            super().remove_child(child)

    def add_placeholder(self, mod):
        placeholder = self.__get_placeholder()

        if placeholder is None:
            placeholder = MrgPlaceholder(visible=True, controller=self)
            self.add(placeholder)

        self.show_child(placeholder)

    def remove_placeholder(self, mod):
        placeholder = self.__get_placeholder()
        if placeholder:
            self.remove_child(placeholder)
